import json
import uuid
import re
from io import BytesIO, StringIO
import xml.etree.cElementTree as ET
from urllib.parse import urlencode, urlparse, urlunparse, parse_qs, ParseResult

from arcgis.gis import GIS
from arcgis import env as _env
from arcgis._impl.common._mixins import PropertyMap
from ._base import BaseOGC


###########################################################################
class WMSLayer(BaseOGC):
    """
    Represents a Web Map Service, which is an OGC web service endpoint.

    ===============     ====================================================================
    **Parameter**        **Description**
    ---------------     --------------------------------------------------------------------
    url                 Required string. The administration URL for the ArcGIS Server.
    ---------------     --------------------------------------------------------------------
    version             Optional String. The version number of the WMS service.  The default is `1.3.0`.
    ---------------     --------------------------------------------------------------------
    gis                 Optional :class:`~arcgis.gis.GIS`. The GIS used to reference the service by. The arcgis.env.active_gis is used if not specified.
    ---------------     --------------------------------------------------------------------
    copyright           Optional String. Describes limitations and usage of the data.
    ---------------     --------------------------------------------------------------------
    scale               Optional Tuple. The min/max scale of the layer where the positions are: (min, max) as float values.
    ---------------     --------------------------------------------------------------------
    opacity             Optional Float.  This value can range between 1 and 0, where 0 is 100 percent transparent and 1 is completely opaque.
    ---------------     --------------------------------------------------------------------
    title               Optional String. The title of the layer used to identify it in places such as the Legend and Layer List widgets.
    ===============     ====================================================================

    """

    _gis = None
    _con = None
    _url = None
    _reader = None
    _cap_reader = None
    _properties = None
    _type = "WMS"

    # ----------------------------------------------------------------------
    def __init__(self, url, version="1.3.0", gis=None, **kwargs):
        super(WMSLayer, self)
        if gis:
            gis = gis
        elif gis is None and _env.active_gis:
            gis = _env.active_gis
        else:
            gis = GIS()
        assert isinstance(gis, GIS)
        self._id = kwargs.pop("id", uuid.uuid4().hex)
        self._version = version
        self._con = gis._con
        self._title = kwargs.pop("title", "WMS Layer")
        self._gis = gis
        if url[-1] == "/":
            url = url[:-1]
        self._url = url
        self._add_token = str(self._con._auth).lower() == "builtin"
        self._opacity = kwargs.pop("opacity", 1)
        self._min_scale, self._max_scale = kwargs.pop("scale", (0, 0))

    # ----------------------------------------------------------------------
    @property
    def properties(self) -> PropertyMap:
        """
        Returns the properties of the Layer.

        :return: PropertyMap
        """
        if self._properties is None:
            if self._add_token:
                url = self._capabilities_url(
                    service_url=self._url, vendor_kwargs={"token": self._con.token}
                )
            else:
                url = self._capabilities_url(service_url=self._url)
            text = self._con.get(url, {}, try_json=False, add_token=False)
            if text.find("Invalid Token") > -1 or text.find("Get Token") > -1:
                url = self._capabilities_url(service_url=self._url)
                text = self._con.get(url, {}, try_json=False, add_token=False)
            elif text.lower().find("<html>") > -1:
                url = self._capabilities_url(service_url=self._url)
                text = self._con.get(url, {}, try_json=False, add_token=False)
            elif text.lower().find("<?xml version=") > -1:
                pass
            else:
                raise Exception("Could not connect to the Web Map Service")
            sss = BytesIO()
            sss.write(text.encode())
            sss.seek(0)
            tree = ET.XML(text=sss.read())
            d = self._xml_to_dictionary(tree)
            self._properties = PropertyMap(d)
        return self._properties

    # ---------------------------------------------------------------------
    @property
    def _extents(self) -> list:
        """list of extents from the service in the form of
        [[minx, miny], [maxx, maxy]] for each entry in the list
        """
        try:
            bboxes = self.properties.WMS_Capabilities.Capability.Layer.BoundingBox
            output = []
            for bbox in bboxes:
                output.append(
                    [
                        [float(bbox["@minx"]), float(bbox["@miny"])],
                        [float(bbox["@maxx"]), float(bbox["@maxy"])],
                    ]
                )
            return output
        except Exception:
            return [
                [[0, 0], [0, 0]],
            ]

    @property
    def _spatial_references(self) -> list:
        try:
            crss = self.properties.WMS_Capabilities.Capability.Layer.CRS
            output = []
            for crs_str in crss:
                output += [int(crs_num) for crs_num in re.findall(r"[0-9]+", crs_str)]
            return output
        except Exception as e:
            return []

    # ----------------------------------------------------------------------
    @property
    def layers(self) -> list:
        """returns the layers of the WMS Layer"""
        try:
            return self.properties.WMS_Capabilities.Capability.Layer.Layer
        except:
            return self.properties.WMS_Capabilities.Capability.Layer

    # ----------------------------------------------------------------------
    def _capabilities_url(self, service_url: str, vendor_kwargs: dict = None) -> str:
        """Return a capabilities url"""
        pieces = urlparse(service_url)
        args = parse_qs(pieces.query)
        if "service" not in args:
            args["service"] = "WMS"
        if "request" not in args:
            args["request"] = "GetCapabilities"
        if "version" not in args:
            args["version"] = self._version
        if vendor_kwargs:
            args.update(vendor_kwargs)
        query = urlencode(args, doseq=True)
        pieces = ParseResult(
            pieces.scheme,
            pieces.netloc,
            pieces.path,
            pieces.params,
            query,
            pieces.fragment,
        )
        return urlunparse(pieces)

    # ----------------------------------------------------------------------
    def _format_tags(self, tag: str) -> str:
        """attempts to format tags by stripping out the {text} from the keys"""
        import re

        regex = r".*\}(.*)"
        matches = re.search(regex, tag)
        if matches:
            return matches.groups()[0]
        return tag

    # ----------------------------------------------------------------------
    def _xml_to_dictionary(self, t) -> dict:
        """converts the xml to a dictionary object (recursivly)"""
        import json
        from collections import defaultdict

        d = {self._format_tags(t.tag): {} if t.attrib else None}
        children = list(t)
        if children:
            dd = defaultdict(list)
            for dc in map(self._xml_to_dictionary, children):
                for k, v in dc.items():
                    dd[self._format_tags(k)].append(v)
            d = {
                self._format_tags(t.tag): {
                    self._format_tags(k): v[0] if len(v) == 1 else v
                    for k, v in dd.items()
                }
            }
        if t.attrib:
            d[self._format_tags(t.tag)].update(
                [("@" + self._format_tags(k), v) for k, v in t.attrib.items()]
            )
        if t.text:
            text = t.text.strip()
            if children or t.attrib:
                if text:
                    d[self._format_tags(t.tag)]["#text"] = text
            else:
                d[self._format_tags(t.tag)] = text
        removals = [
            "{http://www.opengis.net/wmts/1.0}",
            "{http://www.opengis.net/ows/1.1}",
            "{http://www.w3.org/1999/xlink}",
        ]
        d = json.dumps(d)
        for remove in removals:
            d = d.replace(remove, "")
        return json.loads(d)

    # ----------------------------------------------------------------------
    @property
    def _lyr_json(self) -> dict:
        """Represents the MapView's widget JSON format"""
        return {
            "type": self._type,
            "id": self._id,
            "title": self._title or "WMTS Layer",
            "url": self._url,
            "version": self._version,
            "sublayers": [{"name": lyr.Name} for lyr in self.layers],
            "minScale": self.scale[0],
            "maxScale": self.scale[1],
            "opacity": self.opacity,
        }

    @property
    def _operational_layer_json(self) -> dict:
        """Represents the WebMap's JSON format"""
        new_layer = self._lyr_json
        new_layer["layers"] = [
            {"name": subLyr.Name, "title": subLyr.Title} for subLyr in self.layers
        ]
        new_layer["visibleLayers"] = []
        if new_layer["layers"]:
            # Only have the first layer be the visible layer
            new_layer["visibleLayers"].append(new_layer["layers"][0]["name"])
        new_layer["extent"] = self._extents[0]
        new_layer["spatialReferences"] = self._spatial_references
        return new_layer



from requests.auth import AuthBase
from urllib import parse

from ._schain import SupportMultiAuth
from arcgis.auth.tools import parse_url


class EsriNotebookAuth(AuthBase, SupportMultiAuth):
    """authentication for notebook servers Keys"""

    _invalid_token_urls = None
    _token = None
    auth = None

    # ----------------------------------------------------------------------
    def __init__(
        self,
        token: str,
        referer: str = None,
        auth: AuthBase = None,
        **kwargs,
    ):
        self._token = token
        self.auth = auth
        if referer is None:
            self.referer = ""
        else:
            self.referer = referer
        self._no_go_token = set()
        self.verify_cert = kwargs.pop("verify_cert", True)

    # ----------------------------------------------------------------------
    def __str__(self):
        return f"<{self.__class__.__name__}, token={self.token[:5]}...>"

    # ----------------------------------------------------------------------
    def __repr__(self):
        return f"<{self.__class__.__name__}, token={self.token[:5]}...>"

    # ----------------------------------------------------------------------
    @property
    def token(self) -> str:
        """
        Gets/Sets the API token

        :returns: String
        """
        return self._token

    # ----------------------------------------------------------------------
    @token.setter
    def token(self, token: str):
        """Gets/Sets the Notebook Token"""
        if self._token != token:
            self._token = token

    # ----------------------------------------------------------------------
    def add_token(self, r, **kwargs):
        """generates a server token using Portal token"""
        if (
            r.text.lower().find("invalid token") > -1
            or r.text.lower().find("token required") > -1
        ):
            parsed = parse_url(r.url)
            self._no_go_token.add((parsed.scheme, parsed.netloc, parsed.path))
            # Recreate the request without the token
            #
            r.content
            r.raw.release_conn()
            r.request.headers["referer"] = self.referer  # or "http"
            r.request.headers.pop("X-Esri-Authorization", None)
            _r = r.connection.send(r.request, **kwargs)
            _r.headers["referer"] = self.referer  # or "http"
            _r.headers.pop("X-Esri-Authorization", None)
            _r.history.append(r)
            return _r
        return r

    # ----------------------------------------------------------------------
    def __call__(self, r):
        parsed = parse_url(r.url)
        if self.auth:
            self.auth.__call__(r)
        if (
            not "X-Esri-Authorization" in r.headers
            and not (parsed.scheme, parsed.netloc, parsed.path) in self._no_go_token
        ):
            r.headers["X-Esri-Authorization"] = f"Bearer {self.token}"
            r.headers.pop("Referer", None)  # ['Referer'] = "http"  # , self.referer
        r.register_hook("response", self.add_token)
        return r



"""
The ``arcgis.geoenrichment`` module provides access to Business Analyst's 
demographic and jurisdictional areas data through enrichment, standard 
geography queries and reporting. The source for analysis can be either 
local or remote. Using a local source requires ArcGIS Pro with the 
Business Analyst extension and at least one country's data pack to be 
installed. A remote source requires access to a property configured Web 
GIS. A Web GIS can either be ArcGIS Online or an instance of ArcGIS 
Enterprise with Business Analyst.
"""

__all__ = [
    "BufferStudyArea",
    "Country",
    "create_report",
    "enrich",
    "get_countries",
    "service_limits",
    "standard_geography_query",
    "interesting_facts",
]

from .enrichment import (
    BufferStudyArea,
    Country,
    create_report,
    enrich,
    get_countries,
    service_limits,
    standard_geography_query,
    interesting_facts,
)



# Copyright (C) 2019 NVIDIA Corporation. Ting-Chun Wang, Ming-Yu Liu, Jun-Yan Zhu.
# BSD License. All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.

# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.

# THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR ANY PARTICULAR PURPOSE.
# IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL
# DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
# WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING
# OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.


# --------------------------- LICENSE FOR pytorch-CycleGAN-and-pix2pix ----------------
# Copyright (c) 2017, Jun-Yan Zhu and Taesung Park
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.

# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# Based on https://github.com/NVIDIA/pix2pixHD

from torch.autograd import Variable
from fastai.vision import (
    LearnerCallback,
    functools,
    np,
    flatten_model,
    requires_grad,
    SmoothenValue,
    add_metrics,
    F,
)
import torch
from torch import nn


def weights_init(m):
    classname = m.__class__.__name__
    if classname.find("Conv") != -1:
        m.weight.data.normal_(0.0, 0.02)
    elif classname.find("BatchNorm2d") != -1:
        m.weight.data.normal_(1.0, 0.02)
        m.bias.data.fill_(0)


def get_norm_layer(norm_type="instance"):
    if norm_type == "batch":
        norm_layer = functools.partial(nn.BatchNorm2d, affine=True)
    elif norm_type == "instance":
        norm_layer = functools.partial(nn.InstanceNorm2d, affine=False)
    else:
        raise NotImplementedError("normalization layer [%s] is not found" % norm_type)
    return norm_layer


def define_gen(
    input_nc,
    output_nc,
    ngf,
    netG,
    n_downsample_global=3,
    n_blocks_global=9,
    n_local_enhancers=1,
    n_blocks_local=3,
    norm="instance",
    gpu_ids=[],
):
    norm_layer = get_norm_layer(norm_type=norm)
    if netG == "global":
        netG = GlobalGenerator(
            input_nc, output_nc, ngf, n_downsample_global, n_blocks_global, norm_layer
        )
    elif netG == "local":
        netG = LocalEnhancer(
            input_nc,
            output_nc,
            ngf,
            n_downsample_global,
            n_blocks_global,
            n_local_enhancers,
            n_blocks_local,
            norm_layer,
        )
    elif netG == "encoder":
        netG = Encoder(input_nc, output_nc, ngf, n_downsample_global, norm_layer)
    else:
        raise ("generator not implemented!")
    if len(gpu_ids) > 0:
        assert torch.cuda.is_available()
        netG.cuda(gpu_ids[0])
    netG.apply(weights_init)
    return netG


def define_dscr(
    input_nc,
    ndf,
    n_layers_D,
    norm="instance",
    use_sigmoid=False,
    num_D=1,
    getIntermFeat=False,
    gpu_ids=[],
):
    norm_layer = get_norm_layer(norm_type=norm)
    netD = MultiscaleDiscriminator(
        input_nc, ndf, n_layers_D, norm_layer, use_sigmoid, num_D, getIntermFeat
    )
    if len(gpu_ids) > 0:
        assert torch.cuda.is_available()
        netD.cuda(gpu_ids[0])
    netD.apply(weights_init)
    return netD


class GANLoss(nn.Module):
    def __init__(
        self,
        use_lsgan=True,
        target_real_label=1.0,
        target_fake_label=0.0,
        tensor=torch.FloatTensor,
    ):
        super(GANLoss, self).__init__()
        self.real_label = target_real_label
        self.fake_label = target_fake_label
        self.real_label_var = None
        self.fake_label_var = None
        self.Tensor = tensor
        if use_lsgan:
            self.loss = nn.MSELoss()
        else:
            self.loss = nn.BCELoss()

    def get_target_tensor(self, input, target_is_real):
        target_tensor = None
        if target_is_real:
            create_label = (self.real_label_var is None) or (
                self.real_label_var.numel() != input.numel()
            )
            if create_label:
                real_tensor = self.Tensor(input.size()).fill_(self.real_label)
                self.real_label_var = Variable(real_tensor, requires_grad=False)
            target_tensor = self.real_label_var
        else:
            create_label = (self.fake_label_var is None) or (
                self.fake_label_var.numel() != input.numel()
            )
            if create_label:
                fake_tensor = self.Tensor(input.size()).fill_(self.fake_label)
                self.fake_label_var = Variable(fake_tensor, requires_grad=False)
            target_tensor = self.fake_label_var
        return target_tensor

    def __call__(self, input, target_is_real):
        if isinstance(input[0], list):
            loss = 0
            for input_i in input:
                pred = input_i[-1]
                target_tensor = self.get_target_tensor(pred, target_is_real)
                loss += self.loss(pred, target_tensor)
            return loss
        else:
            target_tensor = self.get_target_tensor(input[-1], target_is_real)
            return self.loss(input[-1], target_tensor)


class VGGLoss(nn.Module):
    def __init__(self, gpu_ids):
        super(VGGLoss, self).__init__()
        self.vgg = Vgg19().cuda()
        self.criterion = nn.L1Loss()
        self.weights = [1.0 / 32, 1.0 / 16, 1.0 / 8, 1.0 / 4, 1.0]

    def forward(self, x, y):
        x_vgg, y_vgg = self.vgg(x), self.vgg(y)
        loss = 0
        for i in range(len(x_vgg)):
            loss += self.weights[i] * self.criterion(x_vgg[i], y_vgg[i].detach())
        return loss


##############################################################################
# Generator
##############################################################################


class LocalEnhancer(nn.Module):
    def __init__(
        self,
        input_nc,
        output_nc,
        ngf=32,
        n_downsample_global=3,
        n_blocks_global=9,
        n_local_enhancers=1,
        n_blocks_local=3,
        norm_layer=nn.BatchNorm2d,
        padding_type="reflect",
    ):
        super(LocalEnhancer, self).__init__()
        self.n_local_enhancers = n_local_enhancers

        ###### global generator model #####
        ngf_global = ngf * (2**n_local_enhancers)
        model_global = GlobalGenerator(
            input_nc,
            output_nc,
            ngf_global,
            n_downsample_global,
            n_blocks_global,
            norm_layer,
        ).model
        model_global = [
            model_global[i] for i in range(len(model_global) - 3)
        ]  # get rid of final convolution layers
        self.model = nn.Sequential(*model_global)

        ###### local enhancer layers #####
        for n in range(1, n_local_enhancers + 1):
            ### downsample
            ngf_global = ngf * (2 ** (n_local_enhancers - n))
            model_downsample = [
                nn.ReflectionPad2d(3),
                nn.Conv2d(input_nc, ngf_global, kernel_size=7, padding=0),
                norm_layer(ngf_global),
                nn.ReLU(True),
                nn.Conv2d(
                    ngf_global, ngf_global * 2, kernel_size=3, stride=2, padding=1
                ),
                norm_layer(ngf_global * 2),
                nn.ReLU(True),
            ]
            ### residual blocks
            model_upsample = []
            for i in range(n_blocks_local):
                model_upsample += [
                    ResnetBlock(
                        ngf_global * 2, padding_type=padding_type, norm_layer=norm_layer
                    )
                ]

            ### upsample
            model_upsample += [
                nn.ConvTranspose2d(
                    ngf_global * 2,
                    ngf_global,
                    kernel_size=3,
                    stride=2,
                    padding=1,
                    output_padding=1,
                ),
                norm_layer(ngf_global),
                nn.ReLU(True),
            ]

            ### final convolution
            if n == n_local_enhancers:
                model_upsample += [
                    nn.ReflectionPad2d(3),
                    nn.Conv2d(ngf, output_nc, kernel_size=7, padding=0),
                    nn.Tanh(),
                ]

            setattr(self, "model" + str(n) + "_1", nn.Sequential(*model_downsample))
            setattr(self, "model" + str(n) + "_2", nn.Sequential(*model_upsample))

        self.downsample = nn.AvgPool2d(
            3, stride=2, padding=[1, 1], count_include_pad=False
        )

    def forward(self, input):
        ### create input pyramid
        input_downsampled = [input]
        for i in range(self.n_local_enhancers):
            input_downsampled.append(self.downsample(input_downsampled[-1]))

        ### output at coarest level
        output_prev = self.model(input_downsampled[-1])
        ### build up one layer at a time
        for n_local_enhancers in range(1, self.n_local_enhancers + 1):
            model_downsample = getattr(self, "model" + str(n_local_enhancers) + "_1")
            model_upsample = getattr(self, "model" + str(n_local_enhancers) + "_2")
            input_i = input_downsampled[self.n_local_enhancers - n_local_enhancers]
            output_prev = model_upsample(model_downsample(input_i) + output_prev)
        return output_prev


class GlobalGenerator(nn.Module):
    def __init__(
        self,
        input_nc,
        output_nc,
        ngf=64,
        n_downsampling=3,
        n_blocks=9,
        norm_layer=nn.BatchNorm2d,
        padding_type="reflect",
    ):
        assert n_blocks >= 0
        super(GlobalGenerator, self).__init__()
        activation = nn.ReLU(True)

        model = [
            nn.ReflectionPad2d(3),
            nn.Conv2d(input_nc, ngf, kernel_size=7, padding=0),
            norm_layer(ngf),
            activation,
        ]
        ### downsample
        for i in range(n_downsampling):
            mult = 2**i
            model += [
                nn.Conv2d(
                    ngf * mult, ngf * mult * 2, kernel_size=3, stride=2, padding=1
                ),
                norm_layer(ngf * mult * 2),
                activation,
            ]

        ### resnet blocks
        mult = 2**n_downsampling
        for i in range(n_blocks):
            model += [
                ResnetBlock(
                    ngf * mult,
                    padding_type=padding_type,
                    activation=activation,
                    norm_layer=norm_layer,
                )
            ]

        ### upsample
        for i in range(n_downsampling):
            mult = 2 ** (n_downsampling - i)
            model += [
                nn.ConvTranspose2d(
                    ngf * mult,
                    int(ngf * mult / 2),
                    kernel_size=3,
                    stride=2,
                    padding=1,
                    output_padding=1,
                ),
                norm_layer(int(ngf * mult / 2)),
                activation,
            ]
        model += [
            nn.ReflectionPad2d(3),
            nn.Conv2d(ngf, output_nc, kernel_size=7, padding=0),
            nn.Tanh(),
        ]
        self.model = nn.Sequential(*model)

    def forward(self, input):
        return self.model(input)


# Define a resnet block
class ResnetBlock(nn.Module):
    def __init__(
        self, dim, padding_type, norm_layer, activation=nn.ReLU(True), use_dropout=False
    ):
        super(ResnetBlock, self).__init__()
        self.conv_block = self.build_conv_block(
            dim, padding_type, norm_layer, activation, use_dropout
        )

    def build_conv_block(self, dim, padding_type, norm_layer, activation, use_dropout):
        conv_block = []
        p = 0
        if padding_type == "reflect":
            conv_block += [nn.ReflectionPad2d(1)]
        elif padding_type == "replicate":
            conv_block += [nn.ReplicationPad2d(1)]
        elif padding_type == "zero":
            p = 1
        else:
            raise NotImplementedError("padding [%s] is not implemented" % padding_type)

        conv_block += [
            nn.Conv2d(dim, dim, kernel_size=3, padding=p),
            norm_layer(dim),
            activation,
        ]
        if use_dropout:
            conv_block += [nn.Dropout(0.5)]

        p = 0
        if padding_type == "reflect":
            conv_block += [nn.ReflectionPad2d(1)]
        elif padding_type == "replicate":
            conv_block += [nn.ReplicationPad2d(1)]
        elif padding_type == "zero":
            p = 1
        else:
            raise NotImplementedError("padding [%s] is not implemented" % padding_type)
        conv_block += [nn.Conv2d(dim, dim, kernel_size=3, padding=p), norm_layer(dim)]

        return nn.Sequential(*conv_block)

    def forward(self, x):
        out = x + self.conv_block(x)
        return out


class Encoder(nn.Module):
    def __init__(
        self, input_nc, output_nc, ngf=32, n_downsampling=4, norm_layer=nn.BatchNorm2d
    ):
        super(Encoder, self).__init__()
        self.output_nc = output_nc

        model = [
            nn.ReflectionPad2d(3),
            nn.Conv2d(input_nc, ngf, kernel_size=7, padding=0),
            norm_layer(ngf),
            nn.ReLU(True),
        ]
        ### downsample
        for i in range(n_downsampling):
            mult = 2**i
            model += [
                nn.Conv2d(
                    ngf * mult, ngf * mult * 2, kernel_size=3, stride=2, padding=1
                ),
                norm_layer(ngf * mult * 2),
                nn.ReLU(True),
            ]

        ### upsample
        for i in range(n_downsampling):
            mult = 2 ** (n_downsampling - i)
            model += [
                nn.ConvTranspose2d(
                    ngf * mult,
                    int(ngf * mult / 2),
                    kernel_size=3,
                    stride=2,
                    padding=1,
                    output_padding=1,
                ),
                norm_layer(int(ngf * mult / 2)),
                nn.ReLU(True),
            ]

        model += [
            nn.ReflectionPad2d(3),
            nn.Conv2d(ngf, output_nc, kernel_size=7, padding=0),
            nn.Tanh(),
        ]
        self.model = nn.Sequential(*model)

    def forward(self, input, inst):
        outputs = self.model(input)

        # instance-wise average pooling
        outputs_mean = outputs.clone()
        inst_list = np.unique(inst.cpu().numpy().astype(int))
        for i in inst_list:
            for b in range(input.size()[0]):
                indices = (inst[b : b + 1] == int(i)).nonzero()  # n x 4
                for j in range(self.output_nc):
                    output_ins = outputs[
                        indices[:, 0] + b,
                        indices[:, 1] + j,
                        indices[:, 2],
                        indices[:, 3],
                    ]
                    mean_feat = torch.mean(output_ins).expand_as(output_ins)
                    outputs_mean[
                        indices[:, 0] + b,
                        indices[:, 1] + j,
                        indices[:, 2],
                        indices[:, 3],
                    ] = mean_feat
        return outputs_mean


class MultiscaleDiscriminator(nn.Module):
    def __init__(
        self,
        input_nc,
        ndf=64,
        n_layers=3,
        norm_layer=nn.BatchNorm2d,
        use_sigmoid=False,
        num_D=3,
        getIntermFeat=False,
    ):
        super(MultiscaleDiscriminator, self).__init__()
        self.num_D = num_D
        self.n_layers = n_layers
        self.getIntermFeat = getIntermFeat

        for i in range(num_D):
            netD = NLayerDiscriminator(
                input_nc, ndf, n_layers, norm_layer, use_sigmoid, getIntermFeat
            )
            if getIntermFeat:
                for j in range(n_layers + 2):
                    setattr(
                        self,
                        "scale" + str(i) + "_layer" + str(j),
                        getattr(netD, "model" + str(j)),
                    )
            else:
                setattr(self, "layer" + str(i), netD.model)

        self.downsample = nn.AvgPool2d(
            3, stride=2, padding=[1, 1], count_include_pad=False
        )

    def singleD_forward(self, model, input):
        if self.getIntermFeat:
            result = [input]
            for i in range(len(model)):
                result.append(model[i](result[-1]))
            return result[1:]
        else:
            return [model(input)]

    def forward(self, input):
        num_D = self.num_D
        result = []
        input_downsampled = input
        for i in range(num_D):
            if self.getIntermFeat:
                model = [
                    getattr(self, "scale" + str(num_D - 1 - i) + "_layer" + str(j))
                    for j in range(self.n_layers + 2)
                ]
            else:
                model = getattr(self, "layer" + str(num_D - 1 - i))
            result.append(self.singleD_forward(model, input_downsampled))
            if i != (num_D - 1):
                input_downsampled = self.downsample(input_downsampled)
        return result


# Defines the PatchGAN discriminator with the specified arguments.
class NLayerDiscriminator(nn.Module):
    def __init__(
        self,
        input_nc,
        ndf=64,
        n_layers=3,
        norm_layer=nn.BatchNorm2d,
        use_sigmoid=False,
        getIntermFeat=False,
    ):
        super(NLayerDiscriminator, self).__init__()
        self.getIntermFeat = getIntermFeat
        self.n_layers = n_layers

        kw = 4
        padw = int(np.ceil((kw - 1.0) / 2))
        sequence = [
            [
                nn.Conv2d(input_nc, ndf, kernel_size=kw, stride=2, padding=padw),
                nn.LeakyReLU(0.2, True),
            ]
        ]

        nf = ndf
        for n in range(1, n_layers):
            nf_prev = nf
            nf = min(nf * 2, 512)
            sequence += [
                [
                    nn.Conv2d(nf_prev, nf, kernel_size=kw, stride=2, padding=padw),
                    norm_layer(nf),
                    nn.LeakyReLU(0.2, True),
                ]
            ]

        nf_prev = nf
        nf = min(nf * 2, 512)
        sequence += [
            [
                nn.Conv2d(nf_prev, nf, kernel_size=kw, stride=1, padding=padw),
                norm_layer(nf),
                nn.LeakyReLU(0.2, True),
            ]
        ]

        sequence += [[nn.Conv2d(nf, 1, kernel_size=kw, stride=1, padding=padw)]]

        if use_sigmoid:
            sequence += [[nn.Sigmoid()]]

        if getIntermFeat:
            for n in range(len(sequence)):
                setattr(self, "model" + str(n), nn.Sequential(*sequence[n]))
        else:
            sequence_stream = []
            for n in range(len(sequence)):
                sequence_stream += sequence[n]
            self.model = nn.Sequential(*sequence_stream)

    def forward(self, input):
        if self.getIntermFeat:
            res = [input]
            for n in range(self.n_layers + 2):
                model = getattr(self, "model" + str(n))
                res.append(model(res[-1]))
            return res[1:]
        else:
            return self.model(input)


from torchvision import models


class Vgg19(torch.nn.Module):
    def __init__(self, requires_grad=False):
        super(Vgg19, self).__init__()
        vgg_pretrained_features = models.vgg19(pretrained=True).features
        self.slice1 = torch.nn.Sequential()
        self.slice2 = torch.nn.Sequential()
        self.slice3 = torch.nn.Sequential()
        self.slice4 = torch.nn.Sequential()
        self.slice5 = torch.nn.Sequential()
        for x in range(2):
            self.slice1.add_module(str(x), vgg_pretrained_features[x])
        for x in range(2, 7):
            self.slice2.add_module(str(x), vgg_pretrained_features[x])
        for x in range(7, 12):
            self.slice3.add_module(str(x), vgg_pretrained_features[x])
        for x in range(12, 21):
            self.slice4.add_module(str(x), vgg_pretrained_features[x])
        for x in range(21, 30):
            self.slice5.add_module(str(x), vgg_pretrained_features[x])
        if not requires_grad:
            for param in self.parameters():
                param.requires_grad = False

    def forward(self, X):
        h_relu1 = self.slice1(X)
        h_relu2 = self.slice2(h_relu1)
        h_relu3 = self.slice3(h_relu2)
        h_relu4 = self.slice4(h_relu3)
        h_relu5 = self.slice5(h_relu4)
        out = [h_relu1, h_relu2, h_relu3, h_relu4, h_relu5]
        return out


def encode_input(
    label_map,
    label_nc,
    data_type=32,
    inst_map=None,
    real_image=None,
    feat_map=None,
    infer=False,
):
    if label_nc == 0:
        input_label = label_map.data.cuda()
    else:
        # create one-hot vector for label map
        size = label_map.size()
        oneHot_size = (size[0], label_nc, size[2], size[3])
        input_label = torch.cuda.FloatTensor(torch.Size(oneHot_size)).zero_()
        input_label = input_label.scatter_(1, label_map.data.long().cuda(), 1.0)
        if data_type == 16:
            input_label = input_label.half()

    #             get edges from instance map
    #             if not opt.no_instance:
    #                 inst_map = inst_map.data.cuda()
    #                 edge_map = self.get_edges(inst_map)
    #                 input_label = torch.cat((input_label, edge_map), dim=1)
    input_label = Variable(input_label, volatile=infer)

    # real images for training
    if real_image is not None:
        real_image = Variable(real_image.data.cuda())

    # instance map for feature encoding
    #             if self.use_features:
    #                 # get precomputed feature maps
    #                 if self.opt.load_features:
    #                     feat_map = Variable(feat_map.data.cuda())
    #                 if self.opt.label_feat:
    #                     inst_map = label_map.cuda()

    return input_label, inst_map, real_image, feat_map


class Pix2PixHDModel(nn.Module):
    def __init__(self, label_nc, input_nc, output_nc, gpu_ids, **kwargs):
        super().__init__()
        self.gpu_ids = gpu_ids
        ngf = kwargs.get("n_gen_filters", 64)
        n_downsample_global = kwargs.get("n_downsample_global", 4)
        n_blocks_global = kwargs.get("n_blocks_global", 9)
        n_local_enhancers = kwargs.get("n_local_enhancers", 1)
        n_blocks_local = kwargs.get("n_blocks_local", 3)
        self.lsgan = kwargs.get("lsgan", True)

        ndf = kwargs.get("n_dscr_filters", 64)
        self.n_layers_D = kwargs.get("n_layers_dscr", 3)
        self.num_D = kwargs.get("n_dscr", 2)
        self.feat_loss = kwargs.get("feat_loss", True)
        netG = kwargs.get("gen_network", "local")
        norm = kwargs.get("norm", "instance")
        self.label_nc = label_nc

        self.data_type = 32  # hardcoded
        if self.label_nc:
            input_nc = label_nc

        netG_input_nc = input_nc
        self.G = define_gen(
            netG_input_nc,
            output_nc,
            ngf,
            netG,
            n_downsample_global,
            n_blocks_global,
            n_local_enhancers,
            n_blocks_local,
            norm,
            gpu_ids=self.gpu_ids,
        )

        use_sigmoid = not self.lsgan
        netD_input_nc = input_nc + output_nc
        self.D = define_dscr(
            netD_input_nc,
            ndf,
            self.n_layers_D,
            norm,
            use_sigmoid,
            self.num_D,
            self.feat_loss,
            gpu_ids=self.gpu_ids,
        )
        self.arcgis_results = False

    def set_input(self, input):
        self.input_label, self.real_image = input

    def forward(self, label, image):
        if self.label_nc:
            if self.training:
                label = self.input_label
            else:
                label, _, _, _ = encode_input(label_map=label, label_nc=self.label_nc)

        fake_image = self.G(label)

        return [fake_image]


class Pix2PixHDLoss(nn.Module):
    def __init__(
        self,
        p2p_model: nn.Module,
        vgg_loss=True,
        lambda_feat=10.0,
        l1_loss=True,
        lambda_l1=100.0,
        gpu_ids=[],
    ):
        super().__init__()
        self.gpu_ids = gpu_ids
        self.Tensor = torch.cuda.FloatTensor if self.gpu_ids else torch.Tensor
        self.p2p_model = p2p_model
        self.criterionGAN = GANLoss(use_lsgan=self.p2p_model.lsgan, tensor=self.Tensor)
        self.criterionFeat = torch.nn.L1Loss()
        self.lambda_feat = lambda_feat
        self.vgg_loss = vgg_loss
        if self.vgg_loss:
            self.criterionVGG = VGGLoss(self.gpu_ids)
        self.l1_loss = l1_loss
        self.lambda_l1 = lambda_l1

    def set_input(self, input):
        self.input_label, self.real_image = input

    def discriminate(self, input_label, test_image):
        input_concat = torch.cat((input_label, test_image.detach()), dim=1)
        return self.p2p_model.D.forward(input_concat)

    def forward(self, output, target):
        fake_image = output[0]

        # Real Detection and Loss

        pred_real = self.discriminate(self.input_label, self.real_image)  #####

        # GAN loss (Fake Passability Loss)
        pred_fake = self.p2p_model.D.forward(
            torch.cat((self.input_label, fake_image), dim=1)
        )

        self.loss_G_GAN = self.criterionGAN(pred_fake, True)

        # GAN feature matching loss
        self.loss_G_GAN_Feat = 0
        if self.p2p_model.feat_loss:
            feat_weights = 4.0 / (self.p2p_model.n_layers_D + 1)
            D_weights = 1.0 / self.p2p_model.num_D
            for i in range(self.p2p_model.num_D):
                for j in range(len(pred_fake[i]) - 1):
                    self.loss_G_GAN_Feat += (
                        D_weights
                        * feat_weights
                        * self.criterionFeat(pred_fake[i][j], pred_real[i][j].detach())
                        * self.lambda_feat
                    )
        # VGG feature matching loss
        self.loss_G_VGG = 0
        if self.vgg_loss:
            self.loss_G_VGG = (
                self.criterionVGG(fake_image, self.real_image) * self.lambda_feat
            )

        # l1 loss for MS data
        self.loss_G_l1 = 0
        if self.l1_loss:
            self.loss_G_l1 = (
                torch.mean(F.l1_loss(fake_image, self.real_image))
            ) * self.lambda_l1

        return self.loss_G_GAN + self.loss_G_GAN_Feat + self.loss_G_VGG + self.loss_G_l1


class Pix2PixHDTrainer(LearnerCallback):
    _order = -20

    def _set_trainable(self, D=False):
        gen = not D
        # requires_grad(self.learn.model.G, gen)
        # requires_grad(self.learn.model.D, D)
        if not gen:
            self.opt_D.lr, self.opt_D.mom = self.learn.opt.lr, self.learn.opt.mom
            self.opt_D.wd, self.opt_D.beta = self.learn.opt.wd, self.learn.opt.beta

    def discriminate(self, input_label, test_image):
        input_concat = torch.cat((input_label, test_image.detach()), dim=1)
        return self.D.forward(input_concat)

    def on_train_begin(self, **kwargs):
        self.G = self.learn.model.G
        self.D = self.learn.model.D
        self.criterionGAN = self.learn.loss_func.criterionGAN

        if not getattr(self, "opt_G", None):
            self.opt_G = self.learn.opt.new([nn.Sequential(*flatten_model(self.G))])
        else:
            self.opt_G.lr, self.opt_G.wd = self.opt.lr, self.opt.wd
            self.opt_G.mom, self.opt_G.beta = self.opt.mom, self.opt.beta

        if not getattr(self, "opt_D", None):
            self.opt_D = self.learn.opt.new([nn.Sequential(*flatten_model(self.D))])

        self.learn.opt.opt = self.opt_G.opt

        self._set_trainable()

        self.G_GAN_smter, self.G_GAN_Feat_smter, self.G_VGG_smter, self.G_l1_smter = (
            SmoothenValue(0.98),
            SmoothenValue(0.98),
            SmoothenValue(0.98),
            SmoothenValue(0.98),
        )
        self.D_fake_smter, self.D_real_smter = SmoothenValue(0.98), SmoothenValue(0.98)

        self.data_type = 32
        self.recorder.add_metric_names(["gen_loss", "disc_fake_loss", "disc_real_loss"])

        if self.model.feat_loss:
            self.recorder.add_metric_names(["feat_loss"])

        if self.loss_func.vgg_loss:
            self.recorder.add_metric_names(["VGG_loss"])

        if self.loss_func.l1_loss:
            self.recorder.add_metric_names(["l1_loss"])

    def on_batch_begin(self, last_input, **kwargs):
        last_input[0], _, last_input[1], _ = encode_input(
            last_input[0], label_nc=self.model.label_nc, real_image=last_input[1]
        )
        self.learn.model.set_input(last_input)
        self.learn.loss_func.set_input(last_input)

    def on_backward_begin(self, **kwargs):
        self.G_GAN_smter.add_value(self.loss_func.loss_G_GAN.detach().cpu())
        if self.model.feat_loss:
            self.G_GAN_Feat_smter.add_value(
                self.loss_func.loss_G_GAN_Feat.detach().cpu()
            )
        if self.loss_func.vgg_loss:
            self.G_VGG_smter.add_value(self.loss_func.loss_G_VGG.detach().cpu())

        if self.loss_func.l1_loss:
            self.G_l1_smter.add_value(self.loss_func.loss_G_l1.detach().cpu())

    def on_batch_end(self, last_input, last_output, **kwargs):
        self.G.zero_grad()
        fake_image = last_output[0].detach()
        input_label, real_image = last_input

        self._set_trainable(D=True)
        self.D.zero_grad()

        # Fake Detection and Loss
        self.pred_fake = self.discriminate(input_label, fake_image)
        self.pred_real = self.discriminate(input_label, real_image)

        self.loss_D_fake = self.criterionGAN(self.pred_fake, False)
        self.loss_D_real = self.criterionGAN(self.pred_real, True)

        self.D_fake_smter.add_value(self.loss_D_fake.detach().cpu())
        self.D_real_smter.add_value(self.loss_D_real.detach().cpu())

        loss_D = (self.loss_D_fake + self.loss_D_real) * 0.5

        if self.learn.model.training == True:
            loss_D.backward()

        self.opt_D.step()
        self._set_trainable()

    def on_epoch_end(self, last_metrics, **kwargs):
        smter_list = [self.G_GAN_smter, self.D_fake_smter, self.D_real_smter]

        if self.model.feat_loss:
            smter_list.append(self.G_GAN_Feat_smter)
        if self.loss_func.vgg_loss:
            smter_list.append(self.G_VGG_smter)
        if self.loss_func.l1_loss:
            smter_list.append(self.G_l1_smter)

        return add_metrics(last_metrics, [s.smooth for s in smter_list])



import os
from pathlib import Path
from typing import Optional, Union
import winreg

import numpy as np


def get_child_key_strs(key_path) -> list:
    """
    Get the full path of first generation child keys under the parent key listed.
    Args:
        key_path: Path to the parent key in registry.
    Returns: List of the full path to child keys.
    """
    # open the parent key
    parent_key = winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE, key_path)

    # variables to track progress and store results
    error = False
    counter = 0
    key_list = []

    # while everything is going good
    while not error:
        try:
            # get the child key in the iterated position
            child_key = winreg.EnumKey(parent_key, counter)

            # add the located key to the list
            key_list.append("{}\\{}".format(key_path, child_key))

            # increment the counter
            counter += 1

        # when something blows up...typically because no key is found
        except Exception as e:
            # switch the error flag to true, stopping the iteration
            error = True

    # give the accumulated list back
    return key_list


def get_first_child_key_str(key_path, pattern) -> str:
    """
    Based on the pattern provided, find the key with a matching string in it.
    :param key_path: Full string path to the key.
    :param pattern: Pattern to be located.

    :return: Full path of the first key path matching the provided pattern.
    """
    # get a list of paths to keys under the parent key path provided
    key_list = get_child_key_strs(key_path)

    # iterate the list of key paths
    for key in key_list:
        # if the key matches the pattern
        if key.find(pattern):
            # pass back the provided key path
            return key


def get_ba_country_key_str(
    three_letter_country_code: str, year: Optional[Union[int, str]] = None
) -> str:
    """Lookup the country registry key by three letter country identifier and year."""
    cntry_cd = three_letter_country_code.upper()
    cntry_key_lst = get_child_key_strs(
        r"SOFTWARE\WOW6432Node\Esri\BusinessAnalyst\Datasets"
    )

    key_dict = {
        os.path.basename(k).split("_")[2]: k
        for k in cntry_key_lst
        if os.path.basename(k).split("_")[0] == cntry_cd
    }

    yr_lst = [int(y) for y in key_dict.keys()]
    yr_lst.sort()
    assert len(yr_lst)

    if year:
        year = (
            str(year) if isinstance(year, int) or isinstance(year, np.int64) else year
        )
        assert isinstance(year, str)
        assert len(year) == 4
        cntry_key = key_dict[year]
    else:
        cntry_key = key_dict[str(yr_lst[-1:][0])]
        assert len(
            yr_lst
        ), f"It appears {cntry_cd} {year} is not installed on this machine."

    return cntry_key


def get_ba_key_value(
    locator_key,
    three_letter_country_identifier: str = "USA",
    year: Optional[Union[int, str]] = None,
) -> Path:
    """
    In the Business Analyst key, get the value corresponding to the provided locator key.
    :param locator_key: Locator key.
    :param three_letter_country_identifier: Three letter country identification code.
    :param year: Four digit year describing the vintage of the data.

    :return: Key value.
    """
    # get the registry key string path
    country_key_str = get_ba_country_key_str(three_letter_country_identifier, year)

    # open the key to the current installation of Business Analyst data
    key = winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE, country_key_str)

    # query the value of the locator key
    return winreg.QueryValueEx(key, locator_key)[0]


def get_ba_network_dataset_path(
    three_letter_country_identifier: str, vintage_year: Optional[int] = None
) -> Path:
    """
    Get the path to the transportation network dataset.
    :param three_letter_country_identifier: Three letter country identification code.

    :return: String describing resource location.
    """
    return Path(
        get_ba_key_value(
            "StreetsNetwork", three_letter_country_identifier, vintage_year
        )
    )



from . import hub
from . import workforce
from . import storymap
from . import survey123
from . import tracker
from . import dashboard
from . import expbuilder

from ._url_schemes import build_collector_url
from ._url_schemes import build_field_maps_url
from ._url_schemes import build_explorer_url
from ._url_schemes import build_navigator_url
from ._url_schemes import build_survey123_url
from ._url_schemes import build_tracker_url
from ._url_schemes import build_workforce_url

try:
    from arcgis.widgets._mapview import MapView
except ImportError as e:
    import logging

    log = logging.getLogger()
    import_error = e

    class MapView:
        def __init__(self, *args, **kwargs):
            log.warning(
                "MapView class replaced with a non-functional "
                "placeholder due to the following import error:"
            )
            raise import_error



from typing import AsyncGenerator, Union, Optional, ClassVar
from dataclasses import dataclass

from arcgis.realtime import Velocity
from arcgis.realtime.velocity.feeds._feed_template import _FeedTemplate
from arcgis.realtime.velocity.feeds.geometry import (
    _HasGeometry,
    SingleFieldGeometry,
    XYZGeometry,
)
from arcgis.realtime.velocity.feeds.time import _HasTime, TimeInstant, TimeInterval
from arcgis.realtime.velocity.input.format import (
    EsriJsonFormat,
    GeoJsonFormat,
    JsonFormat,
    DelimitedFormat,
    XMLFormat,
    _format_from_config,
)


@dataclass
class HttpReceiver(_FeedTemplate, _HasTime, _HasGeometry):
    """
    Receive events via a dedicated HTTP endpoint. This data class can be used to define the feed configuration
    and to create the feed.

    ====================      ======================================================================
    **Parameter**              **Description**
    --------------------      ----------------------------------------------------------------------
    label                     String. Unique label for this feed instance.
    --------------------      ----------------------------------------------------------------------
    description               String. Feed description.
    --------------------      ----------------------------------------------------------------------
    authentication_type       String. Authentication type.

                              Options:

                                none or arcgis.
    --------------------      ----------------------------------------------------------------------
    sample_message            String. The sample content to auto-detect the data format.

                              For example:

                                "name,age\\nsam,23"
    ====================      ======================================================================

    =====================     ====================================================================
    **Optional Argument**     **Description**
    =====================     ====================================================================
    data_format               [:class:`~arcgis.realtime.velocity.input.EsriJsonFormat`, :class:`~arcgis.realtime.velocity.input.GeoJsonFormat`, :class:`~arcgis.realtime.velocity.input.DelimitedFormat`, :class:`~arcgis.realtime.velocity.input.JsonFormat`, :class:`~arcgis.realtime.velocity.input.XMLFormat`].
                              An instance that contains the data format
                              configuration for this feed. Configure only allowed formats.
                              If this is not set right during initialization, a format will be
                              auto-detected and set from a sample of the incoming data. This sample
                              will be fetched from the configuration provided so far in the init.
    ---------------------     --------------------------------------------------------------------
    track_id_field            String. Name of the field from the incoming data that should be set as
                              track ID.
    ---------------------     --------------------------------------------------------------------
    geometry                  [:class:`~arcgis.realtime.velocity.feeds.XYZGeometry`, :class:`~arcgis.realtime.velocity.feeds.SingleFieldGeometry`]. An instance of geometry
                              configuration that will be used to create geometry objects from the incoming data.
    ---------------------     --------------------------------------------------------------------
    time                      [:class:`~arcgis.realtime.velocity.feeds.TimeInstant`, :class:`~arcgis.realtime.velocity.feeds.TimeInterval`]. An instance of time configuration that
                              will be used to create time information from the incoming data.
    =====================     ====================================================================

    :return: A data class with Http receiver feed configuration.

    .. code-block:: python

        # Usage Example

        from arcgis.realtime.velocity.feeds import HttpReceiver
        from arcgis.realtime.velocity.http_authentication_type import (
            NoAuth,
            BasicAuth,
            CertificateAuth,
        )

        sample_message="name,age\n
        dan,23"

        http_receiver = HttpReceiver(
            label="feed_name",
            description="feed_description",
            authentication_type="none",
            sample_message=sample_message,
            data_format=None
        )

        # use velocity object to get the FeedsManager instance
        feeds = velocity.feeds

        # use the FeedsManager object to create a feed from this feed configuration
        http_receiver_feed = feeds.create(http_receiver)
        http_receiver_feed.start()
        feeds.items

    """

    # fields that the user sets during init
    # HTTP Poller specific properties
    authentication_type: str
    sample_message: str

    # user can define these properties even after initialization
    data_format: Optional[
        Union[EsriJsonFormat, GeoJsonFormat, JsonFormat, DelimitedFormat, XMLFormat]
    ] = None
    # FeedTemplate properties
    track_id_field: Optional[str] = None
    # HasGeometry properties
    geometry: Optional[Union[XYZGeometry, SingleFieldGeometry]] = None
    # HasTime properties
    time: Optional[Union[TimeInstant, TimeInterval]] = None

    # FeedTemplate properties
    _name: ClassVar[str] = "http-receiver"

    def __post_init__(self):
        if Velocity is None:
            return
        self._util = Velocity._util

        # validation of fields
        if self._util.is_valid(self.label) == False:
            raise ValueError(
                "Label should only contain alpha numeric, _ and space only"
            )
        if self.authentication_type not in ("none", "arcgis"):
            raise ValueError("authentication_type must be either 'none' or 'arcgis'")
        elif not self.sample_message:
            raise ValueError("sample_message must not be empty")

        # generate dictionary of this feed object's properties that will be used to query test-connection and
        # sample-messages Rest endpoint
        feed_properties = self._generate_feed_properties()

        derived_schema = self._util.derive(sample_data=self.sample_message)
        if "schema" in derived_schema and derived_schema["schema"] is not None:
            self._set_fields(derived_schema["schema"])

            # if Format was not specified by user, use the auto-detected format from the sample messages response as this feed object's format.
            if self.data_format is None:
                self.data_format = _format_from_config(derived_schema)

            print(
                "Feature Schema retrieved from the Feed:",
                derived_schema["schema"],
            )

            # initiate actions for each of the following properties if it was set at init
            if self.track_id_field is not None:
                self.set_track_id(self.track_id_field)
            if self.geometry is not None:
                self.set_geometry_config(self.geometry)
            if self.time is not None:
                self.set_time_config(self.time)
        elif ("status", "error") in derived_schema.items():
            english_messages = map(
                lambda message: message["englishMessage"], derived_schema["messages"]
            )
            errors = "\n".join(english_messages)
            print(
                f"Could not derive schema from sample_message because of the following reasons:\n{errors}"
            )
        else:
            print("Unknown error detecting the format from sample_message")

    def _build(self) -> dict:
        feed_configuration = {
            "id": "",
            "label": self.label,
            "description": self.description,
            "feed": {**self._generate_schema_transformation()},
            "properties": {"executable": True},
        }

        feed_properties = self._generate_feed_properties()
        self._dict_deep_merge(feed_configuration["feed"], feed_properties)
        print(feed_configuration)
        return feed_configuration

    def _generate_feed_properties(self) -> dict:
        feed_properties = {
            "name": self._name,
            "properties": {
                f"{self._name}.httpAuthenticationType": self.authentication_type
            },
        }

        if self.data_format is not None:
            format_dict = self.data_format._build()
            self._dict_deep_merge(feed_properties, format_dict)

        return feed_properties



# Copyright 2018-2019 Open-MMLab.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# cascade_rcnn_s101_fpn_syncbn-backbone+head_mstrain-range_1x_coco.py, box AP=46.8

_base_ = "./_base_/models/cascade_rcnn_r50_fpn.py"
norm_cfg = dict(type="GN", num_groups=32, requires_grad=True)
model = dict(
    pretrained="open-mmlab://resnest101",
    backbone=dict(
        type="ResNeSt",
        stem_channels=128,
        depth=101,
        radix=2,
        reduction_factor=4,
        avg_down_stride=True,
        num_stages=4,
        out_indices=(0, 1, 2, 3),
        frozen_stages=1,
        norm_cfg=norm_cfg,
        norm_eval=False,
        style="pytorch",
    ),
    roi_head=dict(
        bbox_head=[
            dict(
                type="Shared4Conv1FCBBoxHead",
                in_channels=256,
                conv_out_channels=256,
                fc_out_channels=1024,
                norm_cfg=norm_cfg,
                roi_feat_size=7,
                num_classes=80,
                bbox_coder=dict(
                    type="DeltaXYWHBBoxCoder",
                    target_means=[0.0, 0.0, 0.0, 0.0],
                    target_stds=[0.1, 0.1, 0.2, 0.2],
                ),
                reg_class_agnostic=True,
                loss_cls=dict(
                    type="CrossEntropyLoss", use_sigmoid=False, loss_weight=1.0
                ),
                loss_bbox=dict(type="SmoothL1Loss", beta=1.0, loss_weight=1.0),
            ),
            dict(
                type="Shared4Conv1FCBBoxHead",
                in_channels=256,
                conv_out_channels=256,
                fc_out_channels=1024,
                norm_cfg=norm_cfg,
                roi_feat_size=7,
                num_classes=80,
                bbox_coder=dict(
                    type="DeltaXYWHBBoxCoder",
                    target_means=[0.0, 0.0, 0.0, 0.0],
                    target_stds=[0.05, 0.05, 0.1, 0.1],
                ),
                reg_class_agnostic=True,
                loss_cls=dict(
                    type="CrossEntropyLoss", use_sigmoid=False, loss_weight=1.0
                ),
                loss_bbox=dict(type="SmoothL1Loss", beta=1.0, loss_weight=1.0),
            ),
            dict(
                type="Shared4Conv1FCBBoxHead",
                in_channels=256,
                conv_out_channels=256,
                fc_out_channels=1024,
                norm_cfg=norm_cfg,
                roi_feat_size=7,
                num_classes=80,
                bbox_coder=dict(
                    type="DeltaXYWHBBoxCoder",
                    target_means=[0.0, 0.0, 0.0, 0.0],
                    target_stds=[0.033, 0.033, 0.067, 0.067],
                ),
                reg_class_agnostic=True,
                loss_cls=dict(
                    type="CrossEntropyLoss", use_sigmoid=False, loss_weight=1.0
                ),
                loss_bbox=dict(type="SmoothL1Loss", beta=1.0, loss_weight=1.0),
            ),
        ],
    ),
)

checkpoint = "http://download.openmmlab.com/mmdetection/v2.0/resnest/cascade_rcnn_s101_fpn_syncbn-backbone%2Bhead_mstrain-range_1x_coco/cascade_rcnn_s101_fpn_syncbn-backbone%2Bhead_mstrain-range_1x_coco_20201005_113242-b9459f8f.pth"



"""
Allows for requests authentication to be chained.

Inspired by: https://github.com/Colin-b/requests_auth

"""
from requests.auth import AuthBase


###########################################################################
class _MultiAuth(AuthBase):
    """Authentication using multiple authentication methods."""

    def __init__(self, *authentication_modes):
        self.authentication_modes = authentication_modes

    def __call__(self, r):
        for authentication_mode in self.authentication_modes:
            authentication_mode.__call__(r)
        return r

    def __add__(self, other):
        if isinstance(other, _MultiAuth):
            return _MultiAuth(*self.authentication_modes, *other.authentication_modes)
        return _MultiAuth(*self.authentication_modes, other)

    def __and__(self, other):
        if isinstance(other, _MultiAuth):
            return _MultiAuth(*self.authentication_modes, *other.authentication_modes)
        return _MultiAuth(*self.authentication_modes, other)


###########################################################################
class SupportMultiAuth:
    """Inherit from this class to be able to use your class with requests_auth provided authentication classes."""

    def __add__(self, other):
        if isinstance(other, _MultiAuth):
            return _MultiAuth(self, *other.authentication_modes)
        return _MultiAuth(self, other)

    def __and__(self, other):
        if isinstance(other, _MultiAuth):
            return _MultiAuth(self, *other.authentication_modes)
        return _MultiAuth(self, other)



"""
Contains the base class that all server object inherit from.
"""
from __future__ import absolute_import
from __future__ import annotations
import json
import functools
from collections import OrderedDict
from urllib.request import HTTPError
from ..._impl._con import Connection
from arcgis.gis import GIS
from arcgis._impl.common._mixins import PropertyMap


###########################################################################
class BaseServer(object):
    """class most server object inherit from"""

    _con = None
    _url = None
    _json_dict = None
    _json = None
    _properties = None

    def __init__(self, url, gis=None, initialize=True, **kwargs):
        """class initializer"""
        if gis is None and "connection" in kwargs:
            connection = kwargs["connection"]
            gis = kwargs.pop("connection", None)
        super(BaseServer, self).__init__()
        self._url = url

        # gis = kwargs.pop('gis', None)
        if gis is None and isinstance(gis, GIS):
            gis = gis._portal.con

        if isinstance(gis, Connection):
            self._con = gis
        elif hasattr(gis, "_con"):
            self._gis = gis._con
        else:
            raise ValueError("gis must be of type SiteConnection")
        if initialize:
            self._init(gis)

    @functools.lru_cache(maxsize=10)
    def _server_version(self) -> list[float]:
        """returns the server version number"""
        params: dict[str, Any] = {"f": "json"}
        url: str = self._url.split("/admin/")[0] + "/admin"
        return [
            int(v) for v in str(self._con.get(url, params)["currentVersion"]).split(".")
        ]

    # ----------------------------------------------------------------------
    def _init(self, connection=None):
        """loads the properties into the class"""
        if connection is None:
            connection = self._con
        params = {"f": "json"}
        try:
            result = connection.get(path=self._url, params=params)
            if isinstance(result, dict):
                self._json_dict = result
                self._properties = PropertyMap(result)
            else:
                self._json_dict = {}
                self._properties = PropertyMap({})
        except HTTPError as err:
            raise RuntimeError(err)
        except:
            self._json_dict = {}
            self._properties = PropertyMap({})

    # ----------------------------------------------------------------------
    def __str__(self):
        return "< %s @ %s >" % (type(self).__name__, self._url)

    # ----------------------------------------------------------------------
    def __repr__(self):
        return "< %s @ %s >" % (type(self).__name__, self._url)

    # ----------------------------------------------------------------------
    @property
    def properties(self):
        """
        returns the object properties
        """
        if self._properties is None:
            self._init()
        return self._properties

    # ----------------------------------------------------------------------
    def __getattr__(self, name):
        """adds dot notation to any class"""
        if self._properties is None:
            self._init()
        try:
            return self._properties.__getitem__(name)
        except:
            for k, v in self._json_dict.items():
                if k.lower() == name.lower():
                    return v
            raise AttributeError(
                "'%s' object has no attribute '%s'" % (type(self).__name__, name)
            )

    # ----------------------------------------------------------------------
    def __getitem__(self, key):
        """helps make object function like a dictionary object"""
        try:
            return self._properties.__getitem__(key)
        except KeyError:
            for k, v in self._json_dict.items():
                if k.lower() == key.lower():
                    return v
            raise AttributeError(
                "'%s' object has no attribute '%s'" % (type(self).__name__, key)
            )
        except:
            raise AttributeError(
                "'%s' object has no attribute '%s'" % (type(self).__name__, key)
            )

    # ----------------------------------------------------------------------
    @property
    def url(self):
        """gets/sets the service url"""
        return self._url

    # ----------------------------------------------------------------------
    @url.setter
    def url(self, value):
        """gets/sets the service url"""
        self._url = value
        self._refresh()

    # ----------------------------------------------------------------------
    def __iter__(self):
        """creates iterable for classes properties"""
        for k, v in self._json_dict.items():
            yield k, v

    # ----------------------------------------------------------------------
    def _refresh(self):
        """reloads all the properties of a given service"""
        self._init()



from ._machine_learning import MLModel, raise_data_exception
import os
import shutil
import tempfile
import random
import json
import pickle
import warnings
import math
import shutil
import os
import time
from pathlib import Path
import traceback
import arcgis
from arcgis.features import FeatureLayer
from ._codetemplate import ml_raster_prf

HAS_AUTO_ML_DEPS = True
import_exception = None

try:
    from ._arcgis_model import ArcGISModel, _raise_fastai_import_error
    from arcgis.learn._fairlearn._fairlearn import calculate_metrics
    from arcgis.learn._utils.tabular_data import (
        TabularDataObject,
        add_h3,
        _extract_embeddings,
    )
    from arcgis.learn._utils.common import _get_emd_path
    from arcgis.learn._utils.utils import arcpy_localization_helper
    import pickle
    from sklearn.preprocessing import normalize

    HAS_FASTAI = True
except:
    import_exception = traceback.format_exc()
    HAS_FASTAI = False

try:
    import sklearn
    from sklearn import *
    from sklearn import preprocessing
    import numpy as np
    import pandas as pd
    from sklearn.pipeline import make_pipeline
    from sklearn.compose import make_column_transformer
    from sklearn.impute import SimpleImputer
    from sklearn.preprocessing import (
        Normalizer,
        LabelEncoder,
        MinMaxScaler,
        StandardScaler,
        OrdinalEncoder,
    )
except Exception as e:
    import_exception = "\n".join(
        traceback.format_exception(type(e), e, e.__traceback__)
    )
    HAS_AUTO_ML_DEPS = False

HAS_FAST_PROGRESS = True
try:
    from fastprogress.fastprogress import progress_bar
except Exception as e:
    import_exception = "\n".join(
        traceback.format_exception(type(e), e, e.__traceback__)
    )
    HAS_FAST_PROGRESS = False

_PROTOCOL_LEVEL = 2


class AutoML(object):
    """
    Automates the process of model selection, training and hyperparameter tuning of
    machine learning models within a specified time limit. Based upon
    MLJar(https://github.com/mljar/mljar-supervised/) and scikit-learn.

    Note that automated machine learning support is provided only for supervised learning.
    Refer https://supervised.mljar.com/

    =====================   ===========================================
    **Parameter**            **Description**
    ---------------------   -------------------------------------------
    data                    Required TabularDataObject. Returned data object from
                            :meth:`~arcgis.learn.prepare_tabulardata` function.
    ---------------------   -------------------------------------------
    total_time_limit        Optional Int. The total time limit in seconds for
                            AutoML training.
                            Default is 3600 (1 Hr)
    ---------------------   -------------------------------------------
    mode                    Optional Str.
                            Can be {Basic, Intermediate, Advanced}. This parameter defines
                            the goal of AutoML and how intensive the AutoML search will be.

                            Basic : To to be used when the user wants to explain and understand the data. Uses 75%/25% train/test split. Uses the following models: Baseline, Linear, Decision Tree, Random Trees, XGBoost, Neural Network, and Ensemble. Has full explanations in reports: learning curves, importance  plots, and SHAP plots.
                            Intermediate : To be used when the user wants to train a model that will be used in real-life use cases. Uses 5-fold CV (Cross-Validation). Uses the following models: Linear, Random Trees, LightGBM, XGBoost, CatBoost, Neural Network, and Ensemble. Has learning curves and importance plots in reports.

                            Advanced : To be used for machine learning competitions (maximum performance). Uses 10-fold CV (Cross-Validation). Uses the following models: Decision Tree, Random Trees, Extra Trees, XGBoost, CatBoost, Neural Network, Nearest Neighbors, Ensemble, and Stacking.It has only learning curves in the reports. Default is Basic
    ---------------------   -------------------------------------------
    algorithms              Optional. List of str.
                            The list of algorithms that will be used in the training. The algorithms can be:
                            Linear, Decision Tree, Random Trees, Extra Trees, LightGBM, Xgboost, Neural Network
    ---------------------   -------------------------------------------
    eval_metric             Optional  Str. The metric to be used to compare models.
                            Possible values are:
                            For binary classification - logloss (default), auc, f1, average_precision,
                            accuracy.
                            For multiclass classification - logloss (default), f1, accuracy
                            For regression - rmse (default), mse, mae, r2, mape, spearman, pearson

                            Note - If there are only 2 unique values in the target, then
                            binary classification is performed,
                            If number of unique values in the target is between 2 and 20 (included), then
                            multiclass classification is performed,
                            In all other cases, regression is performed on the dataset.
    ---------------------   -------------------------------------------
    n_jobs                  Optional. Int.
                            Number of CPU cores to be used. By default, it is set to 1.Set it
                            to -1 to use all the cores.
    =====================   ===========================================

    **kwargs**

    =======================   ===========================================
    sensitive_variables       Optional. List of strings.
                              Variables in the feature class/dataframe which are sensitive and prone to model bias.
                              Ex - ['sex','race'] or ['nationality']
    -----------------------   -------------------------------------------
    fairness_metric           Optional. String.
                              Name of fairness metric based on which fairness optimization should be done on the evaluated models.
                              Available metrics for binary classification are 'demographic_parity_difference' , 'demographic_parity_ratio',
                              'equalized_odds_difference', 'equalized_odds_ratio'.
                              'demographic_parity_ratio' is the default.
                              Available metrics for regression are 'group_loss_ratio' (Default) and 'group_loss_difference'.
    -----------------------   -------------------------------------------
    fairness_threshold        Optional. Float.
                              Required when the chosen metric is group_loss_difference
                              The threshold value for fairness metric. Default values are as follows:
                              - for `demographic_parity_difference` the metric value should be below 0.25,
                              - for `demographic_parity_ratio` the metric value should be above 0.8,
                              - for `equalized_odds_difference` the metric value should be below 0.25,
                              - for `equalized_odds_ratio` the metric value should be above 0.8.
                              - for `group_loss_ratio` the metric value should be above 0.8.
                              - for `group_loss_difference` the metric value should be below 0.25,
    -----------------------   -------------------------------------------
    privileged_groups         Optional. List.
                              List of previleged groups in the sensitive attribute.
                              For example, in binary classification task, a privileged group is the one with the highest selection rate.
                              Example value: [{"sex": "Male"}]
    -----------------------   -------------------------------------------
    underprivileged_groups    Optional. List.
                              List of underprivileged groups in the sensitive attribute.
                              For example, in binary classification task, an underprivileged group
                              is the one with the lowest selection rate.
                              Example value: [{"sex": "Female"}]
    =======================   ===========================================

    :return: :class:`~arcgis.learn.AutoML` Object
    """

    def __init__(
        self,
        data=None,
        total_time_limit=3600,
        mode="Basic",
        algorithms=None,
        eval_metric="auto",
        n_jobs=1,
        ml_task="auto",
        **kwargs,
    ):
        try:
            import platform

            if platform.system() == "Linux":
                message = """
                        Please enable tensorflow by setting the required environment variable 'ARCGIS_ENABLE_TF_BACKEND' to '1' before importing arcgis
                        \n for example the following code block needs to be executed before importing arcgis
                        \n\n`import os; os.environ['ARCGIS_ENABLE_TF_BACKEND'] = '1'`
                        """
                print(message)
            from supervised.automl import AutoML as base_AutoML
        except Exception as e:
            import_exception = "\n".join(
                traceback.format_exception(type(e), e, e.__traceback__)
            )
            _raise_fastai_import_error(import_exception=import_exception)

        if not HAS_AUTO_ML_DEPS:
            _raise_fastai_import_error(import_exception=import_exception)

        self._data = data
        if isinstance(self._data._dependent_variable, list):
            self._data._dependent_variable = self._data._dependent_variable[0]

        if getattr(self._data, "_is_unsupervised", False):
            raise Exception(
                "Auto ML feature is currently only available for Supervised learning."
            )
        if getattr(self._data, "_is_not_empty", False):
            if (len(data._training_indexes) < 20) & (
                eval_metric in ["r2", "rmse", "mse", "mape", "spearman", "pearson"]
            ):
                warnings.warn(
                    "The eval metric you have passed, is not valid for a classification usecase. If the use case is regression, then ensure that your dataset has atleast 22 records"
                )
                return

        self._code = ml_raster_prf
        if algorithms:
            algorithms = algorithms
        else:
            algorithms = [
                "Linear",
                "Decision Tree",
                "Random Trees",
                "Extra Trees",
                "LightGBM",
                "Xgboost",
                "Neural Network",
            ]
        try:
            algorithms = [
                "Random Trees" if x == "Random Forest" else x for x in algorithms
            ]  # To handle backward compatibility with RF
        except:
            pass

        if getattr(self._data, "_is_not_empty", True):
            (
                self._training_data,
                self._training_labels,
                self._validation_data,
                self._validation_labels,
            ) = self._data._ml_data

            self._all_data_df = self._data._dataframe[
                self._data._continuous_variables
                + self._data._categorical_variables
                + self._data._embedding_variables
            ]
            self._all_data_df = self._impute_missing_values(data=self._all_data_df)
            self._all_labels = self._data._dataframe[
                self._data._dependent_variable
            ]  # .values
            self._validation_data_df = pd.DataFrame(
                self._validation_data,
                columns=self._data._continuous_variables
                + self._data._categorical_variables
                + self._data._embedding_variables,
            )
            if ml_task == "auto":
                ml_task = self.get_ml_task(self._all_labels)
            if ml_task == "text":
                msg = arcpy_localization_helper(
                    "Dependent variable has more than 200 unique values more than half of the total records are unique, hence there is not enough information to train a model",
                    260154,
                    "ERROR",
                )
                exit(260146)
            if (mode == "Explain") or (mode == "Basic"):
                explain_level = 2
                zone_list = ["zone3_id", "zone4_id", "zone5_id", "zone6_id", "zone7_id"]
                # try:
                #    self._all_data_df = self._all_data_df.drop(columns=zone_list)
                #    self._data._continuous_variables = [x for x in self._data._continuous_variables if x not in zone_list]
                #    self._data._categorical_variables = [x for x in self._data._categorical_variables if x not in zone_list]
                # except:
                #    pass
            else:
                explain_level = 0  # Setting explain level to 0 in case of Perform and Compete as EDA seems to be creating memory issues

            # Mapping and conversion of old mode names to new
            api_modes = ["Explain", "Perform", "Compete"]
            tool_modes = ["Basic", "Intermediate", "Advanced"]
            if mode in tool_modes:
                mode = api_modes[tool_modes.index(mode)]
            elif mode in api_modes:
                mode = mode
            else:
                mode = "Explain"

            if self._data._embedding_variables and mode != "Compete":
                warnings.warn(
                    "AutoML will be trained in Advanced/Compete mode when text or Image variables are used in model training."
                )
                mode = "Compete"

            try:
                import arcpy

                result_path = tempfile.mkdtemp(dir=arcpy.env.scratchFolder)
            except:
                result_path = tempfile.mkdtemp(dir=tempfile.gettempdir())

            self._sensitive_variables = kwargs.get("sensitive_variables", None)
            self._fairness_metric = kwargs.get("fairness_metric", "auto")
            self._fairness_threshold = kwargs.get("fairness_threshold", "auto")
            self._privileged_groups = kwargs.get("privileged_groups", [])
            self._underprivileged_groups = kwargs.get("underprivileged_groups", [])

            for grp in self._underprivileged_groups:
                for key in grp:
                    val = grp[key]
                    if val == "":
                        self._underprivileged_groups = []

            if (
                self._fairness_metric == "group_loss_difference"
                and self._fairness_threshold == "auto"
            ):
                warnings.warn(
                    "Fairness Threshold value is required to be passed when the chosen fairness metric is group_loss_difference."
                )
                # exit()

            if self._fairness_metric == "equalised_odds_ratio":
                self._fairness_metric = "equalized_odds_ratio"
            if self._fairness_metric == "equalised_odds_difference":
                self._fairness_metric = "equalized_odds_difference"

            self._model = base_AutoML(
                results_path=result_path,
                mode=mode,
                ml_task=ml_task,
                algorithms=algorithms,
                total_time_limit=total_time_limit,
                golden_features=False,
                explain_level=explain_level,
                eval_metric=eval_metric,
                n_jobs=n_jobs,
                kmeans_features=False,
                fairness_metric=self._fairness_metric,
                fairness_threshold=self._fairness_threshold,
                privileged_groups=self._privileged_groups,
                underprivileged_groups=self._underprivileged_groups,
            )
        else:
            result_path = self._data.path
            self._model = base_AutoML(results_path=result_path)
            self._model._results_path = self._data.path

    def get_ml_task(self, all_labels):
        try:
            if isinstance(all_labels[0], str):
                unique = np.unique(all_labels, return_counts=False)
                if len(unique) == 2:
                    return "binary_classification"
                elif len(unique) > 200 and len(unique) > int(0.5 * all_labels.shape[0]):
                    return "text"
                else:
                    return "multiclass_classification"
            else:
                return "auto"
        except:
            return "auto"

    def _impute_missing_values(self, data=None):
        original_dtype = data.dtypes
        numerical_transformer = make_pipeline(SimpleImputer(strategy="median"))

        categorical_transformer = make_pipeline(SimpleImputer(strategy="constant"))

        _procs = make_column_transformer(
            (numerical_transformer, self._data._continuous_variables),
            (categorical_transformer, self._data._categorical_variables),
            (numerical_transformer, self._data._embedding_variables),
        )
        if data is None:
            data = self._all_data_df
        try:
            processed_data = _procs.fit_transform(data)
            processed_data_df = pd.DataFrame(
                processed_data, columns=data.columns.values.tolist()
            )
            processed_data_df = processed_data_df.astype(original_dtype)
        except:
            processed_data_df = data
        return processed_data_df

    def fit(self, sample_weight=None):
        """
        Fits the AutoML model.
        """
        if getattr(self._data, "_is_not_empty", True):
            if isinstance(self._all_labels[0], int):
                self._all_labels = self._all_labels.astype(np.int32)
            elif isinstance(self._all_labels[0], float):
                self._all_labels = self._all_labels.astype(np.float)
            if self._sensitive_variables:
                sensitive_features = self._all_data_df[
                    self._sensitive_variables
                ].astype("category")
            else:
                sensitive_features = None
            try:
                self._model.fit(
                    self._all_data_df,
                    self._all_labels,
                    sample_weight=sample_weight,
                    sensitive_features=sensitive_features,
                )
            except:
                msg = arcpy_localization_helper(
                    "The desired models could not be trained using the input data provided.",
                    260150,
                    "ERROR",
                )
                exit()
        else:
            raise Exception("Fit can be called only with data.")
        # self.save()
        print(
            "All the evaluated models are saved in the path ",
            os.path.abspath(self._model._get_results_path()),
        )

    def show_results(self, rows=5):
        """
        Shows sample results for the model.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        rows                    Optional number of rows. By default, 5 rows
                                are displayed.
        =====================   ===========================================
        :return:
            dataframe
        """
        if getattr(self._data, "is_not_empty", True) == False:
            raise Exception(
                "This method is not available when the model is initiated for prediction"
            )
        if (
            not self._data._is_unsupervised
            and (self._validation_data is None or self._validation_labels is None)
        ) or (self._data._is_unsupervised and self._validation_data is None):
            raise_data_exception()

        min_size = len(self._validation_data)

        if rows < min_size:
            min_size = rows

        # sample_batch = random.sample(self._data._validation_indexes, min_size)
        sample_batch = random.sample(range(len(self._validation_data)), min_size)
        validation_data_batch = self._validation_data.take(sample_batch, axis=0)
        # validation_data_batch_df = pd.DataFrame(validation_data_batch,
        # columns=self._data._continuous_variables + self._data._categorical_variables)
        sample_indexes = [self._data._validation_indexes[i] for i in sample_batch]
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            output_labels = self._predict(validation_data_batch)
        pd.options.mode.chained_assignment = None
        if self._data._is_classification:
            df = self._data._dataframe.loc[sample_indexes]
        else:
            df = self._data._dataframe.iloc[
                sample_indexes
            ]  # .loc[sample_batch]#.reset_index(drop=True).loc[sample_batch].reset_index(drop=True)

        if self._data._dependent_variable:
            df[self._data._dependent_variable + "_results"] = output_labels
        else:
            df["prediction_results"] = output_labels

        return df.sort_index()

    def score(self):
        """
        :return:
            output from AutoML's model.score(), R2 score in case of regression and Accuracy in case of classification.
        """
        col_type = str(self._validation_labels.dtype)
        val_labels = self._validation_labels
        if col_type == "object":
            if isinstance(val_labels[0], float):
                val_labels = val_labels.astype(float)
            elif isinstance(val_labels[0], int):
                val_labels = val_labels.astype(int)
            else:
                val_labels = self._validation_labels
        val_labels = self._validation_labels.astype(int)
        if getattr(self._data, "_is_not_empty", True):
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", UserWarning)
                return self._model.score(self._validation_data_df, val_labels)
        else:
            raise Exception(
                "This method is not available when the model is initiated for prediction"
            )

    def fairness_score(
        self,
        sensitive_feature,
        fairness_metrics=None,
        visualize=False,
    ):
        """
        Shows sample results for the model.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        sensitive_feature       Column name of the protected class.
        ---------------------   -------------------------------------------
        fairness_metrics        Allowed list of fairness metrics. List can
                                have any of the metrics from the list below.
                                Multiple metrics can be passed in the list.
                                 1. For Binary classification
                                    [
                                     "equalized_odds_difference",
                                     "demographic_parity_difference",
                                     "equalized_odds_ratio",
                                     "demographic_parity_ratio"
                                    ]
                                 2. for Regression
                                    [
                                    "MAE",
                                    "MSE",
                                    "RMSE",
                                    "MAPE"
                                    ]
                                 Metric should be one of the values mentioned in
                                 the list.
                                This method is not yet supported for multiclass
                                classification.
        ---------------------   -------------------------------------------
        visualize               A boolean value to visualize plot of metrics
        =====================   ===========================================
        :return: tuple/dataframe
        """
        if self._data._is_classification:
            validation_indexes = self._data._validation_indexes
        else:
            validation_indexes = self._data._dataframe.sample(
                n=round(0.1 * len(self._data._dataframe)),
                replace=False,
                random_state=42,
            ).index.to_list()
        self.sensitive_feature_series = self._validation_data_df.loc[
            :, [sensitive_feature]
        ]
        if self._sensitive_variables:
            return "Since AutoML was trained with fairness mitigation, the fairness score can be obtained by running the report() method."

        if not getattr(self._data, "_is_not_empty", True):
            raise Exception(
                "This method is not available when the model is initiated for prediction"
            )

        y_true = self._data._dataframe.loc[validation_indexes][
            self._data._dependent_variable
        ]
        y_true = y_true.reset_index(drop=True)
        y_pred = self.predict(
            self._data._dataframe.loc[validation_indexes], prediction_type="dataframe"
        )
        y_pred = y_pred["prediction_results"]
        y_pred = y_pred.reset_index(drop=True)

        if self._data._is_classification:
            le_1 = LabelEncoder()
            le_1.fit(y_true)
            y_true = le_1.transform(y_true)
            y_pred = le_1.transform(y_pred)

        return calculate_metrics(
            self._data._is_classification,
            self._data,
            y_true,
            y_pred,
            self.sensitive_feature_series,
            sensitive_feature,
            fairness_metrics,
            visualize,
        )

    def report(self):
        """
        :return:
            a report of the different models trained by AutoML along with their performance.
        """
        main_readme_html = os.path.join(self._model._results_path, "README.html")
        warnings.warn(
            "In case the report html is not rendered appropriately in the notebook, the same can be found in the path "
            "" + main_readme_html
        )
        return self._model.report()

    def predict_proba(self):
        """
        :return:
            output from AutoML's model.predict_proba() with prediction probability for the training data
        """
        if (self._data._is_classification == "classification") or (
            self._data._is_classification == True
        ):
            if getattr(self._data, "_is_not_empty", False):
                raise Exception(
                    "This method is not available when the model is initiated for prediction"
                )
            else:
                cols = (
                    self._data._continuous_variables + self._data._categorical_variables
                )
                data_df = pd.DataFrame(self._data._ml_data[0], columns=cols)
                return self._model.predict_proba(data_df)
        else:
            raise Exception("This method is applicable only for classification models.")

    def copy_and_overwrite(self, from_path, to_path):
        dest_dir = os.path.join(to_path, os.path.basename(from_path))
        if os.path.exists(dest_dir):
            shutil.rmtree(dest_dir)
        shutil.copytree(from_path, dest_dir)

    def _copy_reports(self, src_dir, dest):
        extensions = [".html", ".png", ".svg"]

        for root, dirs, files in os.walk(src_dir):
            for file in files:
                for extension in extensions:
                    if file.endswith(extension):
                        folder_name = os.path.basename(root)
                        src = os.path.join(root, file)
                        dest_folder = dest
                        if not folder_name.startswith("tmp"):
                            dest_folder = os.path.join(dest, folder_name)
                            if not os.path.isdir(dest_folder):
                                os.mkdir(dest_folder)
                        dest_folder = os.path.join(dest_folder, file)
                        shutil.copy2(src, dest_folder)

    def save(self, path):
        """
        Saves the model in the path specified. Creates an Esri Model and a dlpk.
        Uses pickle to save the model and transforms.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        path                    Path of the directory where the model should be saved.
        =====================   ===========================================
        :return:
            path
        """
        if getattr(self._data, "_is_not_empty", True) == False:
            raise Exception(
                "This method is not available when the model is initiated for prediction"
            )
        # Required files to be copied to new path
        files_required = [
            "data_info.json",
            "ldb_performance.png",
            "ldb_performance_boxplot.png",
            "params.json",
            "progress.json",
            "README.md",
            "drop_features.json",
            "model_explainer.sav",
        ]
        required_model_folders = []  # List of folders that are to be copied to new path
        base_file_name = os.path.basename(self._model._get_results_path())
        result_path = os.path.abspath(self._model._get_results_path())

        save_model_path = os.path.abspath(path)
        if not os.path.exists(save_model_path):
            os.makedirs(save_model_path)

        with open(Path(save_model_path) / "ArcGISImageClassifier.py", "w") as f:
            f.write(self._code)

        MLModel._save_encoders(
            self._data._encoder_mapping, save_model_path, base_file_name
        )

        if self._data._procs:
            MLModel._save_transforms(self._data._procs, save_model_path, base_file_name)

        try:
            self._save_explainer(save_model_path)
        except:
            print("Explainer could not be saved")

        self._write_emd(save_model_path, base_file_name)
        if (self._model._best_model._name == "Ensemble") or (
            self._model._best_model._name == "Ensemble_Stacked"
        ):
            model_map = self._model._best_model.models_map
            required_model_folders.append(os.path.join(result_path, "Ensemble"))
            for i in self._model._best_model.selected_models:
                # print(i['model'])
                sub_path = list(model_map.keys())[
                    list(model_map.values()).index(i["model"])
                ]
                final_path = os.path.join(result_path, sub_path)
                required_model_folders.append(final_path)
        else:
            final_path = os.path.join(result_path, self._model._best_model._name)
            required_model_folders.append(final_path)

        for folder in required_model_folders:
            # copyfolder(folder,dest)
            try:
                self.copy_and_overwrite(folder, save_model_path)
            except:
                print(
                    "It looks like the model has been already been saved once. Unable to save at a different location again"
                )
                return

        for file in files_required:
            abs_file_path = os.path.join(result_path, file)
            dest_file = os.path.join(save_model_path, os.path.basename(file))
            if os.path.isfile(abs_file_path):
                shutil.copyfile(abs_file_path, dest_file)

        # Copies reports if present
        try:
            self._copy_reports(result_path, save_model_path)
            copy_success = True
        except:
            copy_success = False
        if copy_success:
            shutil.rmtree(result_path)
        # Creates dlpk

        from ._arcgis_model import _create_zip

        _create_zip(Path(save_model_path).name, str(save_model_path))

        print("Model has been saved in the path", save_model_path)
        return save_model_path

    def _save_explainer(self, path):
        import shap

        if self._model._get_ml_task() == "regression":
            explainer = shap.KernelExplainer(
                self._shap_predict, shap.sample(self._all_data_df.values, 500)
            )
        else:
            explainer = shap.KernelExplainer(
                self._shap_predict,
                shap.sample(self._all_data_df.values, 500),
                link="logit",
            )
        filename = os.path.join(path, "model_explainer.sav")
        pickle.dump(explainer, open(filename, "wb"))

    def _write_emd(self, path, base_file_name):
        def convert(o):
            import numpy

            if isinstance(o, numpy.int64):
                return int(o)

        emd_file = os.path.join(path, base_file_name + ".emd")
        emd_params = {}
        emd_params["version"] = str(sklearn.__version__)
        # if not self._data._is_unsupervised:
        # emd_params["score"] = self.score()
        emd_params["_is_classification"] = (
            "classification" if self._data._is_classification else "regression"
        )
        emd_params["ModelName"] = "AutoML"
        emd_params["ResultsPath"] = self._model._results_path
        # emd_params['ModelFile'] = base_file_name + '.pkl'
        # emd_params['ModelParameters'] = self._model.get_params()
        emd_params["categorical_variables"] = self._data._categorical_variables

        if self._data._dependent_variable:
            emd_params["dependent_variable"] = self._data._dependent_variable

        emd_params["continuous_variables"] = self._data._continuous_variables
        emd_params["text_variables"] = self._data._text_variables
        if self._data._image_variables:
            emd_params["image_variables"] = self._data._image_variables
        emd_params["embedding_variables"] = self._data._embedding_variables
        if self._data._feature_field_variables:
            emd_params["_feature_field_variables"] = self._data._feature_field_variables
        if self._data._raster_field_variables:
            emd_params["_raster_field_variables"] = self._data._raster_field_variables

        emd_params["Framework"] = "arcgis.learn.models._inferencing"
        emd_params["ModelConfiguration"] = "_auto_ml"
        emd_params["InferenceFunction"] = "ArcGISImageClassifier.py"

        if self._model._get_ml_task() != "regression":
            try:
                emd_params["dependent_variable_unique"] = (
                    self._data._dataframe[self._data._dependent_variable]
                    .unique()
                    .tolist()
                )
            except:
                emd_params["dependent_variable_unique"] = []

        with open(emd_file, "w", encoding="utf-8") as f:
            f.write(json.dumps(emd_params, indent=4, default=convert))

    @classmethod
    def from_model(cls, emd_path):
        """
        Creates an `AutoML Model` Object from an Esri Model Definition (EMD) file.
        The model object created can only be used for inference on a new dataset
        and cannot be retrained.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        emd_path                Required string. Path to Esri Model Definition
                                file.
        =====================   ===========================================

        :return:
            :class:`~arcgis.learn.AutoML` Object
        """
        if not HAS_FASTAI:
            _raise_fastai_import_error(import_exception=import_exception)
        emd_path_orig = Path(emd_path)
        emd_path = _get_emd_path(emd_path)
        if not HAS_AUTO_ML_DEPS:
            _raise_fastai_import_error(import_exception=import_exception)

        if not os.path.exists(emd_path):
            raise Exception("Invalid data path.")

        with open(emd_path, "r") as f:
            emd = json.loads(f.read())

        categorical_variables = emd["categorical_variables"]
        dependent_variable = emd.get("dependent_variable", None)
        continuous_variables = emd["continuous_variables"]
        text_variables = emd.get("text_variables", None)
        image_variables = emd.get("image_variables", None)
        embedding_variables = emd.get("embedding_variables", None)

        if emd["version"] != str(sklearn.__version__):
            warnings.warn(
                f"Sklearn version has changed. Model Trained using version {emd['version']}"
            )

        _is_classification = True
        if emd["_is_classification"] != "classification":
            _is_classification = False

        encoder_mapping = None
        if categorical_variables:
            encoder_path = os.path.join(
                os.path.dirname(emd_path),
                os.path.basename(emd_path).split(".")[0] + "_encoders.pkl",
            )
            if os.path.exists(encoder_path):
                with open(encoder_path, "rb") as f:
                    encoder_mapping = pickle.loads(f.read())

        column_transformer = None
        transforms_path = os.path.join(
            os.path.dirname(emd_path),
            os.path.basename(emd_path).split(".")[0] + "_transforms.pkl",
        )
        if os.path.exists(transforms_path):
            with open(transforms_path, "rb") as f:
                column_transformer = pickle.loads(f.read())

        empty_data = TabularDataObject._empty(
            categorical_variables,
            continuous_variables,
            dependent_variable,
            encoder_mapping,
            column_transformer,
            text_variables=text_variables,
            image_variables=image_variables,
            embedding_variables=embedding_variables,
        )
        empty_data._is_classification = _is_classification
        if _is_classification:
            try:
                empty_data.unique_var_list = emd["dependent_variable_unique"]
            except:
                empty_data.unique_var_list = None
        empty_data._is_not_empty = False
        # empty_data.path = emd["ResultsPath"]
        empty_data.path = emd_path.parent
        try:
            empty_data.explainer_path = os.path.join(
                os.path.dirname(emd_path),
                "model_explainer.sav",
            )
        except:
            empty_data.explainer_path = None
        return cls(data=empty_data)

    def _predict(self, data):
        data_df = pd.DataFrame(
            data,
            columns=self._data._continuous_variables
            + self._data._categorical_variables
            + self._data._embedding_variables,
        )
        data_df = self._impute_missing_values(data=data_df)
        return self._model.predict(data_df)

    def _shap_predict(self, data):
        data_df = pd.DataFrame(
            data,
            columns=self._data._continuous_variables
            + self._data._categorical_variables,
        )
        data_df = self._impute_missing_values(data=data_df)
        if self._model._get_ml_task() == "regression":
            return self._model.predict(data_df)
        else:
            return self._model.predict_proba(data_df)

    def _predict_all(self, data):
        data_df = pd.DataFrame(
            data,
            columns=self._data._continuous_variables
            + self._data._categorical_variables,
        )
        data_df = self._impute_missing_values(data=data_df)
        return self._model.predict_all(data_df)

    def _predict_proba(self, data):
        data_df = pd.DataFrame(
            data,
            columns=self._data._continuous_variables
            + self._data._categorical_variables,
        )
        data_df = self._impute_missing_values(data=data_df)
        return self._model.predict_proba(data_df)

    def predict(
        self,
        input_features=None,
        explanatory_rasters=None,
        datefield=None,
        distance_features=None,
        output_layer_name="Prediction Layer",
        gis=None,
        prediction_type="features",
        output_raster_path=None,
        match_field_names=None,
        cell_sizes=[3, 4, 5, 6, 7],
        confidence=True,
        get_local_explanations=False,
        **kwargs,
    ):
        """

        Predict on data from feature layer, dataframe and or raster data.

        =================================   =========================================================================
        **Parameter**                        **Description**
        ---------------------------------   -------------------------------------------------------------------------
        input_features                      Optional :class:`~arcgis.features.FeatureLayer` or spatial dataframe. Required if prediction_type='features'.
                                            Contains features with location and
                                            some or all fields required to infer the dependent variable value.
        ---------------------------------   -------------------------------------------------------------------------
        explanatory_rasters                 Optional list. Required if prediction_type='raster'.
                                            Contains a list of raster objects containing
                                            some or all fields required to infer the dependent variable value.
        ---------------------------------   -------------------------------------------------------------------------
        datefield                           Optional string. Field name from feature layer
                                            that contains the date, time for the input features.
                                            Same as :meth:`~arcgis.learn.prepare_tabulardata`.
        ---------------------------------   -------------------------------------------------------------------------
        cell_sizes                          Size of H3 cells (specified as H3 resolution) for spatially
                                            aggregating input features and passing in the cell ids as additional
                                            explanatory variables to the model. If a spatial dataframe is passed
                                            as input_features, ensure that the spatial reference is 4326,
                                            and the geometry type is Point. Not applicable when explanatory_rasters
                                            are provided.
        ---------------------------------   -------------------------------------------------------------------------
        distance_features                   Optional List of :class:`~arcgis.features.FeatureLayer` objects.
                                            These layers are used for calculation of field "NEAR_DIST_1",
                                            "NEAR_DIST_2" etc in the output dataframe.
                                            These fields contain the nearest feature distance
                                            from the input_features.
                                            Same as :meth:`~arcgis.learn.prepare_tabulardata` .
        ---------------------------------   -------------------------------------------------------------------------
        output_layer_name                   Optional string. Used for publishing the output layer.
        ---------------------------------   -------------------------------------------------------------------------
        gis                                 Optional :class:`~arcgis.gis.GIS`  Object. Used for publishing the item.
                                            If not specified then active gis user is taken.
        ---------------------------------   -------------------------------------------------------------------------
        prediction_type                     Optional String.
                                            Set 'features' or 'dataframe' to make output feature layer predictions.
                                            With this feature_layer argument is required.

                                            Set 'raster', to make prediction raster.
                                            With this rasters must be specified.
        ---------------------------------   -------------------------------------------------------------------------
        output_raster_path                  Optional path.
                                            Required when prediction_type='raster', saves
                                            the output raster to this path.
        ---------------------------------   -------------------------------------------------------------------------
        match_field_names                   Optional dictionary.
                                            Specify mapping of field names from prediction set
                                            to training set.
                                            For example:

                                            |    {
                                            |        "Field_Name_1": "Field_1",
                                            |        "Field_Name_2": "Field_2"
                                            |   }
        ---------------------------------   -------------------------------------------------------------------------
        confidence                          Optional Bool.
                                            Set confidence to True to get prediction confidence for classification
                                            use cases.Default is True.
        =================================   =========================================================================

        :return:
            :class:`~arcgis.features.FeatureLayer` if prediction_type='features', dataframe for prediction_type='dataframe' else creates an output raster.

        """

        rasters = explanatory_rasters if explanatory_rasters else []
        if prediction_type in ["features", "dataframe"]:
            if input_features is None:
                raise Exception("Feature Layer required for predict_features=True")

            gis = gis if gis else arcgis.env.active_gis
            return self._predict_features(
                input_features,
                rasters,
                datefield,
                cell_sizes,
                distance_features,
                output_layer_name,
                gis,
                match_field_names,
                prediction_type,
                confidence,
                get_local_explanations,
                **kwargs,
            )
        else:
            if not rasters:
                raise Exception("Rasters required for predict_features=False")

            if not output_raster_path:
                raise Exception(
                    "Please specify output_raster_folder_path to save the output."
                )

            return self._predict_rasters(
                output_raster_path, rasters, match_field_names, confidence
            )

    def _get_normalised_shap_values(self, processed_numpy):
        # try:
        filename_expl = self._data.explainer_path
        load_explainer = pickle.load(open(filename_expl, "rb"))
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            shap_values = load_explainer.shap_values(processed_numpy, nsamples=100)
        shap_values_normalised = []
        if isinstance(shap_values, np.ndarray):
            shap_values_normalised = normalize(shap_values, axis=1, norm="l1")
        else:
            for val in shap_values:
                normed_matrix = normalize(val, axis=1, norm="l1")
                shap_values_normalised.append(normed_matrix)
        return shap_values_normalised

    def _predict_features(
        self,
        input_features,
        rasters=None,
        datefield=None,
        cell_sizes=[3, 4, 5, 6, 7],
        distance_feature_layers=None,
        output_name="Prediction Layer",
        gis=None,
        match_field_names=None,
        prediction_type="features",
        confidence=False,
        get_local_explanations=False,
        **kwargs,
    ):
        dataframe_complete = False
        attachment_list = kwargs.get("image_attach_list", None)
        if isinstance(input_features, FeatureLayer):
            try:
                import arcpy
                import spatial_reference_helper

                ex = input_features.properties["extent"]
                data_source_spatial_reference = arcpy.SpatialReference(
                    ex["spatialReference"].get(
                        "wkid", ex["spatialReference"].get("latestWkid")
                    )
                )
                extent = arcpy.Extent(
                    ex["xmin"],
                    ex["ymin"],
                    ex["xmax"],
                    ex["ymax"],
                    spatial_reference=data_source_spatial_reference,
                )
                transformation = spatial_reference_helper.get_datum_transformation(
                    data_source_spatial_reference, arcpy.SpatialReference(4326), extent
                )
            except:
                transformation = None
            if cell_sizes and not rasters:
                dataframe = input_features.query(
                    out_sr=4326, datum_transformation=transformation
                ).sdf
                dataframe = add_h3(dataframe, cell_sizes)
            else:
                dataframe = input_features.query().sdf

            if attachment_list:
                dataframe["Images"] = attachment_list
        elif (
            hasattr(input_features, "dataSource")
            or str(input_features).endswith(".shp")
            or isinstance(input_features, tuple)
        ):
            dataframe, index_data = TabularDataObject._sdf_gptool_workflow(
                input_features,
                distance_feature_layers,
                rasters,
                index_field=None,
                is_table_obj=False,
            )
            if cell_sizes and not rasters:
                dataframe = add_h3(dataframe, cell_sizes)
            if attachment_list:
                dataframe["Images"] = attachment_list
            dataframe_complete = True
            self._data._text_variables = self._data._text_variables or []
            self._data._image_variables = self._data._image_variables or []
            if len(self._data._text_variables + self._data._image_variables) > 0:
                dataframe, new_embd_cols = _extract_embeddings(
                    self._data._text_variables, self._data._image_variables, dataframe
                )
        elif hasattr(input_features, "value"):
            dataframe, index_data = TabularDataObject._sdf_gptool_workflow(
                input_features,
                distance_feature_layers,
                rasters,
                index_field=None,
                is_table_obj=True,
            )
            dataframe_complete = True
        else:
            dataframe = input_features.copy()

        self._data._text_variables = self._data._text_variables or []
        self._data._image_variables = self._data._image_variables or []

        fields_needed = (
            self._data._categorical_variables
            + self._data._continuous_variables
            + self._data._text_variables
            + self._data._image_variables
        )
        fields_needed_without_embeddings = (
            self._data._categorical_variables + self._data._continuous_variables
        )
        distance_feature_layers = (
            distance_feature_layers if distance_feature_layers else []
        )
        continuous_variables = self._data._continuous_variables
        non_categorical_variables = (
            self._data._continuous_variables
            + self._data._text_variables
            + self._data._image_variables
        )

        columns = dataframe.columns
        if dataframe_complete:
            processed_dataframe = dataframe
        else:
            feature_layer_columns = []
            for column in columns:
                column_name = column
                categorical = False

                if column_name in fields_needed:
                    if column_name in self._data._text_variables:
                        categorical = "text"
                    elif column_name in self._data._image_variables:
                        categorical = "image"
                    elif column_name not in continuous_variables:
                        categorical = True
                    else:
                        pass
                elif match_field_names and match_field_names.get(column_name):
                    if match_field_names.get(column_name) in self._data._text_variables:
                        categorical = "text"
                    elif (
                        match_field_names.get(column_name)
                        in self._data._image_variables
                    ):
                        categorical = "image"
                    elif match_field_names.get(column_name) not in continuous_variables:
                        categorical = True
                    else:
                        pass
                else:
                    continue

                feature_layer_columns.append((column_name, categorical))

            raster_columns = []
            if rasters:
                for raster in rasters:
                    column_name = raster.name
                    categorical = False
                    if column_name in fields_needed:
                        if column_name not in continuous_variables:
                            categorical = True
                    elif match_field_names and match_field_names.get(column_name):
                        column_name = match_field_names.get(column_name)
                        if column_name not in continuous_variables:
                            categorical = True
                    else:
                        continue

                    raster_columns.append((raster, categorical))

            with warnings.catch_warnings():
                if not HAS_FASTAI:
                    _raise_fastai_import_error(import_exception=import_exception)
                warnings.simplefilter("ignore")
                (
                    processed_dataframe,
                    fields_mapping,
                ) = TabularDataObject._prepare_dataframe_from_features(
                    input_features,
                    self._data._dependent_variable,
                    feature_layer_columns,
                    raster_columns,
                    datefield,
                    cell_sizes,
                    distance_feature_layers,
                )

        if match_field_names:
            try:
                list_of_train_fields = []
                for key, value in match_field_names.items():
                    if (not key == value) and (len(str(key)) > 0):
                        list_of_train_fields.append(value)
                processed_dataframe = processed_dataframe.drop(
                    list_of_train_fields, axis=1, errors="ignore"
                )
            except:
                pass
            processed_dataframe.rename(columns=match_field_names, inplace=True)

        for field in fields_needed_without_embeddings:
            if field not in processed_dataframe.columns:
                msg = arcpy_localization_helper(
                    "Data on which prediction in needed does not have the fields the model was trained on",
                    260152,
                    "ERROR",
                )
                exit()

        for column in processed_dataframe.columns:
            if column not in fields_needed:
                if "emb_" not in column:
                    processed_dataframe = processed_dataframe.drop(column, axis=1)

        processed_numpy = processed_dataframe[
            self._data._continuous_variables
            + self._data._categorical_variables
            + self._data._embedding_variables
        ]
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            predictions = self._predict(processed_numpy)

        if get_local_explanations and self._data._embedding_variables:
            get_local_explanations = False
            warnings.warn(
                "Local explanations cannot be generated when text or image variables are used to train the model."
            )

        if get_local_explanations:
            shap_values_normalised = self._get_normalised_shap_values(processed_numpy)
        shap_df = pd.DataFrame()
        if confidence and self._model._ml_task in [
            "multiclass_classification",
            "binary_classification",
        ]:
            dataframe["prediction_results"] = predictions
            try:
                prediction_confidence = np.amax(
                    self._predict_proba(processed_numpy), axis=1
                )
                dataframe["prediction_confidence"] = prediction_confidence
            except:
                pass

            if get_local_explanations:
                try:
                    labels = self._data.unique_var_list
                    get_element_id = lambda x: labels.index(x)
                    index_id = dataframe["prediction_results"].map(get_element_id)

                    list_for_df = []
                    for index_df, row in dataframe.iterrows():
                        index = index_id[index_df]
                        shap_val_list = shap_values_normalised[index][index_df]
                        list_for_df.append(shap_val_list)
                    shap_df = pd.DataFrame(
                        list_for_df,
                        columns=[
                            i + "_imp"
                            for i in (
                                self._data._continuous_variables
                                + self._data._categorical_variables
                            )
                        ],
                    )
                except:
                    pass
        else:
            dataframe["prediction_results"] = predictions
            if get_local_explanations:
                try:
                    shap_df = pd.DataFrame(
                        shap_values_normalised,
                        columns=[
                            i + "_imp"
                            for i in (
                                self._data._continuous_variables
                                + self._data._categorical_variables
                            )
                        ],
                    )
                except:
                    pass
        dataframe_merged = pd.concat([dataframe, shap_df.abs()], axis=1)
        dataframe_merged = dataframe_merged.filter(regex="^(?!emb_)")

        if prediction_type == "dataframe":
            return dataframe_merged

        return dataframe_merged.spatial.to_featurelayer(output_name, gis)

    def _raster_sr(self, raster):
        try:
            return raster._engine_obj._raster.spatialReference
        except:
            try:
                import arcpy

                return arcpy.SpatialReference(raster.extent["spatialReference"]["wkid"])
            except:
                try:
                    import arcpy

                    return arcpy.SpatialReference(
                        raster.extent["spatialReference"]["wkt"]
                    )
                except:
                    msg = arcpy_localization_helper(
                        "One or more input rasters do not have a valid spatial reference.",
                        517,
                        "ERROR",
                    )

    def _predict_rasters(
        self, output_folder_path, rasters, match_field_names=None, confidence=False
    ):
        if not os.path.exists(os.path.dirname(output_folder_path)):
            raise Exception("Output directory doesn't exist")

        if os.path.exists(output_folder_path):
            raise Exception("Output Folder already exists")

        try:
            import arcpy
        except:
            raise Exception("This function requires arcpy.")

        try:
            import numpy as np
        except:
            raise Exception("This function requires numpy.")

        try:
            import pandas as pd
        except:
            raise Exception("This function requires pandas.")

        if not HAS_FAST_PROGRESS:
            raise Exception("This function requires fastprogress.")

        fields_needed = (
            self._data._categorical_variables + self._data._continuous_variables
        )

        cached_sr = arcpy.env.outputCoordinateSystem
        arcpy.env.outputCoordinateSystem = self._raster_sr(rasters[0])

        xmin = rasters[0].extent["xmin"]
        xmax = rasters[0].extent["xmax"]
        ymin = rasters[0].extent["ymin"]
        ymax = rasters[0].extent["ymax"]
        min_cell_size_x = rasters[0].mean_cell_width
        min_cell_size_y = rasters[0].mean_cell_height

        default_sr = self._raster_sr(rasters[0])

        for raster in rasters:
            point_upper_left = arcpy.PointGeometry(
                arcpy.Point(raster.extent["xmin"], raster.extent["ymax"]),
                self._raster_sr(raster),
            ).projectAs(default_sr)

            point_lower_right = arcpy.PointGeometry(
                arcpy.Point(raster.extent["xmax"], raster.extent["ymin"]),
                self._raster_sr(raster),
            ).projectAs(default_sr)

            cell_extent = arcpy.Extent(
                raster.extent["xmin"],
                raster.extent["ymin"],
                raster.extent["xmin"] + raster.mean_cell_width,
                raster.extent["ymin"] + raster.mean_cell_height,
                spatial_reference=self._raster_sr(raster),
            ).projectAs(default_sr)

            cx, cy = (
                abs(cell_extent.XMax - cell_extent.XMin),
                abs(cell_extent.YMax - cell_extent.YMin),
            )

            if xmin > point_upper_left.firstPoint.X:
                xmin = point_upper_left.firstPoint.X
            if ymax < point_upper_left.firstPoint.Y:
                ymax = point_upper_left.firstPoint.Y
            if xmax < point_lower_right.firstPoint.X:
                xmax = point_lower_right.firstPoint.X
            if ymin > point_lower_right.firstPoint.Y:
                ymin = point_lower_right.firstPoint.Y

            if min_cell_size_x < cx:
                min_cell_size_x = cx

            if min_cell_size_y < cy:
                min_cell_size_y = cy

        max_raster_columns = int(abs(math.ceil((xmax - xmin) / min_cell_size_x)))
        max_raster_rows = int(abs(math.ceil((ymax - ymin) / min_cell_size_y)))
        point_upper = arcpy.PointGeometry(arcpy.Point(xmin, ymax), default_sr)
        point_lower = arcpy.PointGeometry(arcpy.Point(xmax, ymin), default_sr)

        cell_extent = arcpy.Extent(
            xmin,
            ymin,
            xmin + min_cell_size_x,
            ymin + min_cell_size_y,
            spatial_reference=default_sr,
        )

        raster_data = {}
        for raster in rasters:
            field_name = raster.name

            point_upper_translated = point_upper.projectAs(self._raster_sr(raster))
            cell_extent_translated = cell_extent.projectAs(self._raster_sr(raster))

            if field_name not in fields_needed:
                if match_field_names and match_field_names.get(raster.name):
                    field_name = match_field_names.get(raster.name)

            ccxx, ccyy = (
                abs(cell_extent_translated.XMax - cell_extent_translated.XMin),
                abs(cell_extent_translated.YMax - cell_extent_translated.YMin),
            )

            raster_read = raster.read(
                origin_coordinate=(
                    point_upper_translated.firstPoint.X,
                    point_upper_translated.firstPoint.Y,
                ),
                ncols=max_raster_columns,
                nrows=max_raster_rows,
                cell_size=(ccxx, ccyy),
            )

            for row in range(max_raster_rows):
                for column in range(max_raster_columns):
                    values = raster_read[row][column]
                    index = 0
                    for value in values:
                        key = field_name
                        if index != 0:
                            key = key + f"_{index}"
                        if not raster_data.get(key):
                            raster_data[key] = []
                        index = index + 1
                        raster_data[key].append(value)

        for field in fields_needed:
            if (field not in list(raster_data.keys())) and (
                match_field_names and match_field_names.get(field, None) is None
            ):
                msg = arcpy_localization_helper(
                    "Data on which prediction is needed does not have the fields the model was trained on",
                    260152,
                    "ERROR",
                )
                exit()

        processed_data = []

        length_values = len(raster_data[list(raster_data.keys())[0]])
        for i in range(length_values):
            processed_row = []
            for raster_name in sorted(raster_data.keys()):
                processed_row.append(raster_data[raster_name][i])
            processed_data.append(processed_row)

        processed_df = pd.DataFrame(
            data=np.array(processed_data), columns=sorted(raster_data)
        )

        processed_numpy = self._data._process_data(processed_df, fit=False)

        predictions = self._predict(processed_numpy)
        if confidence and self._model._ml_task in [
            "multiclass_classification",
            "binary_classification",
        ]:
            prediction_confidence = np.amax(
                self._predict_proba(processed_numpy), axis=1
            )

        if isinstance(predictions[0], str):
            processed_df["predictions"] = predictions
            le = preprocessing.LabelEncoder()
            le.fit(processed_df["predictions"])
            le_name_mapping = dict(zip(le.classes_, le.transform(le.classes_)))
            processed_df["predictions"] = le.transform(processed_df["predictions"])

            predictions = np.array(
                processed_df["predictions"].values.reshape(
                    [max_raster_rows, max_raster_columns]
                ),
                dtype=np.uint8,
            )
        else:
            predictions = np.array(
                predictions.reshape([max_raster_rows, max_raster_columns]),
                dtype="float64",
            )

        processed_raster = arcpy.NumPyArrayToRaster(
            predictions,
            arcpy.Point(xmin, ymin),
            x_cell_size=min_cell_size_x,
            y_cell_size=min_cell_size_y,
        )
        if isinstance(predictions[0], str):
            arcpy.management.BuildRasterAttributeTable(processed_raster)
            class_map = le_name_mapping
            "class_map=" + str(class_map)
            arcpy.management.CalculateField(
                processed_raster,
                "Class",
                expression="class_map.get(!Value!)",
                expression_type="PYTHON3",
                code_block="class_map=" + str(class_map),
            )
        processed_raster.save(output_folder_path)
        arcpy.env.outputCoordinateSystem = cached_sr

        return True



from __future__ import annotations
from ._base import _BaseKube
from typing import Dict, Any, Optional


###########################################################################
class Mode(_BaseKube):
    _url = None

    @property
    def read_only(self) -> bool:
        """
        Returns if the Kubernetes site is in read only mode.

        :return: bool

        """
        return self.properties

    # ----------------------------------------------------------------------
    def update(
        self, read_only: bool, description: Optional[str] = None
    ) -> Dict[str, str]:
        """
        Updates the site's mode to set it in read only

        ==================     ====================================================================
        **Parameter**           **Description**
        ------------------     --------------------------------------------------------------------
        read_only              Required Boolean.  If True, the organization will be in read only mode.  False it is in write mode.
        ------------------     --------------------------------------------------------------------
        description            Optional String. Sets a custom message to be displayed whenever an attempt to modify or update content or site settings is made through the API. If no custom message is provided, a default response is used.
        ==================     ====================================================================

        :return: Boolean. True if successful else False.


        """
        url = f"{self._url}/update"
        params = {"isReadOnly": read_only}
        if description:
            params["message"] = description
        res = self._con.post(path=url, params=params)
        if "success" in res:
            return res["success"]
        return res



"""A setuptools based setup module.
See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import setup
from setuptools import find_packages
from setuptools.dist import Distribution
from setuptools.command.develop import develop as _develop
from setuptools.command.install import install as _install
from setuptools.command.egg_info import egg_info as _egg_info

# To use a consistent encoding
from codecs import open
from os import path
import sys
from glob import glob
from subprocess import check_output, CalledProcessError, STDOUT
import atexit
import logging
import site

import logging

log = logging.getLogger()
log.setLevel(logging.DEBUG)
handler = logging.StreamHandler(sys.stdout)
handler.setLevel(logging.DEBUG)
log.addHandler(handler)


here = path.abspath(path.dirname(__file__))


class BinaryDistribution(Distribution):
    """Distribution which always forces a binary package with platform name"""

    def has_ext_modules(foo):
        return True


def _get_rel_site_packages_dir():
    for sitepackages in site.getsitepackages():
        try:
            res = "lib" + sitepackages.split("lib")[1] + "/arcgis/gis/_impl"
            if res:
                return res
        except Exception:
            pass


# Conda uses this setup file, but we want to suppress some functionality
if "--conda-install-mode" in sys.argv:
    sys.argv.remove("--conda-install-mode")
    conda_install_mode = True
else:
    conda_install_mode = False

if conda_install_mode:
    # conda handles its own dependencies, so don't specify any pip-dependencies
    dependencies = []
else:
    dependencies = [
        "pillow",
        "urllib3>=1.21.1,<3",
        "cachetools",
        "lxml",
        "notebook",
        "cryptography",
        "ipywidgets >=7,<8",
        "widgetsnbextension >=3",
        "jupyter-client <=6.1.12",
        "pandas >=2.0.0,<3",
        "numpy >=1.21.6",
        "matplotlib",
        "keyring >=23.3.0",
        "pylerc",
        "ujson >=3",
        "jupyterlab",
        "python-certifi-win32;python_version<'3.10'",
        "truststore>=0.7.0;python_version>'3.9'",
        'pywin32 >=223;platform_system=="Windows"',
        "pyshp >=2",
        "geomet",
        "requests >=2.27.1,<3",
        "requests-oauthlib",
        "requests_toolbelt",
        "pyspnego >=0.8.0",
        "requests-kerberos",
        "requests-gssapi",
        "dask >=2023.3.2",
        "matplotlib-inline",
    ]


def _post_install():
    """This function will run after 'pip install' finishes. It has 2 parts:
    1) activate the notebook map widget, equivalent of running these cmds:
        - jupyter nbextension install --py --sys-prefix arcgis
        - jupyter nbextension enable --py --sys-prefix arcgis
        - jupyter nbextension enable --py --sys-prefix widgetsnbextension
    2) If the O.S. is Mac OSX, run the OpenSSL workaround as described in
       this issue: https://bugs.python.org/issue28150, equivalent of running
       '/Applications/Python X.X/Install Certificates.command' cmd
    """
    if conda_install_mode:
        # Don't run any post installation methods for conda installs
        return

    # 1) activate the notebook map widget
    try:
        import notebook.nbextensions as nbext
        import arcgis

        activate_map_widget = True
    except Exception as e:
        log.exception(
            "arcgis/notebook packages don't appear to be installed: "
            "map widget not activated, may not work. The rest of "
            "install is unaffected by this. Exception caught: "
        )
        log.exception(e)
        activate_map_widget = False

    if activate_map_widget:
        log.warning("Attempting to activate map widget...")
        print("Attempting to activate map widget...")
        try:
            log.warning(
                nbext.install_nbextension_python("arcgis", sys_prefix=True, logger=log)
            )

            log.warning(
                nbext.enable_nbextension_python("arcgis", sys_prefix=True, logger=log)
            )

            log.warning(
                nbext.enable_nbextension_python(
                    "widgetsnbextension", sys_prefix=True, logger=log
                )
            )

        except Exception as e:
            print(f"Activating the widget failed {e}")
            log.exception("Activating map widget failed: Continuing install..")
            log.exception(e)

    # 2) If the OS is Mac OSX, run the OpenSSL workaround
    platform_is_osx = sys.platform == "darwin"
    if not platform_is_osx:
        return
    for potential_cert_script in glob("/Applications/Python*/*"):
        if "Install Certificates.command" in potential_cert_script:
            try:
                cmd_output = check_output(potential_cert_script, stderr=STDOUT)
                log.warning(
                    "OpenSSL workaround for OSX completed successfully. "
                    "See https://bugs.python.org/issue28150 for info. "
                    "Output: {}".format(cmd_output.decode("utf-8"))
                )
            except Exception:
                log.exception(
                    "OpenSSL workaround for OSX did not complete "
                    "successfully. This may or may not allow secure SSL "
                    "to work. See https://bugs.python.org/issue28150. "
                )


# Each of these classes represent the different modes that pip install
# can go into, and what logic can be run after pip install finishes
class develop(_develop):
    """Post-installation logic to run for development mode"""

    def run(self):
        self.execute(_post_install, (), msg="Running post-install...")
        super().run()


class install(_install):
    """Post-installation logic to run for installation mode"""

    def run(self):
        self.execute(_post_install, (), msg="Running post-install...")
        super().run()


class egg_info(_egg_info):
    """Post-installation logic to run for 'egg_info' mode"""

    def run(self):
        self.execute(_post_install, (), msg="Running post-install...")
        super().run()


# Read the description.md file
try:
    description_md_file = open("pypi_long_description.md", "r")
    long_description = description_md_file.read()
    description_md_file.close()
except:
    long_description = "ArcGIS API for Python"

# Assemble the `data_files` list of all non-python files
data_files = [
    (
        "share/jupyter/nbextensions/arcgis",
        [
            "arcgis/widgets/js/dist/extension.js",
            "arcgis/widgets/js/dist/arcgis-map-ipywidget.js",
            "arcgis/widgets/js/dist/arcgis-map-ipywidget.js.map",
            "arcgis/apps/workforce/_store/resources/default-project-thumbnail.png",
        ],
    ),
]


def get_version():
    """gets the version from environment variable or sets via manually setting"""
    MAJOR = "2"
    MINOR = "2"
    try:
        import os

        def __path(filename):
            return os.path.join(os.path.dirname(__file__), filename)

        MICRO = "0"
        if os.path.exists(__path("build.info")):
            MICRO = open(__path("build.info")).read().strip()
    except:
        MICRO = "0"
    return f"{MAJOR}.{MINOR}.{MICRO}.2"


kwargs = {
    "name": "arcgis",
    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    "version": get_version(),
    "description": "ArcGIS API for Python",
    "long_description": long_description,
    "long_description_content_type": "text/markdown",
    # The project's main homepage.
    "url": "https://developers.arcgis.com/python/",
    # Author details
    "author": "Esri",
    "author_email": "python@esri.com",
    # Choose your license
    "license": "Esri Master License Agreement (MLA) - http://www.esri.com/LEGAL/pdfs/mla_e204_e300/english.pdf",
    "platforms": ["Windows", "Linux", "Solaris", "Mac OS-X", "Unix"],
    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    "classifiers": [
        # How mature is this project? Common values are
        "Development Status :: 5 - Production/Stable",
        # Topics
        "Topic :: Scientific/Engineering :: GIS",
        "Topic :: Internet :: WWW/HTTP :: Site Management",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Scientific/Engineering :: Visualization",
        "Topic :: Software Development :: Libraries :: Python Modules",
        # Indicate who your project is intended for
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        # Frameworks
        "Framework :: IPython",
        "Framework :: Jupyter",
        # OS
        "Operating System :: OS Independent",
        # Pick your license as you wish (should match "license" above)
        "License :: Other/Proprietary License",
        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
    # What does your project relate to?
    "keywords": "gis arcgis geographic spatial spatial-data "
    "spatial-data-analysis spatial-analysis data-science maps "
    "mapping web-mapping python native-development",
    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    # packages=find_packages(exclude=['contrib', 'docs', 'tests']),
    # Alternatively, if you want to distribute just a my_module.py, uncomment
    # this:
    "packages": find_packages(),
    "python_requires": ">=3.9, <3.12",
    "include_package_data": True,
    "data_files": data_files,
    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    # `setup_dependencies` are all dependencies that need to be in the env
    # BEFORE setup() is called. `install_dependencies` are dependencies
    # that can be put in the env AFTER setup() is finished. See bottom of this
    # file, any unhandled exception will clear `setup_requires` and try again
    # It was thought `setup_requires` needed `notebook` and other deps to
    # activate the map widget, but it caused build errors on windows with
    # required VS C++ distrib (even for --no-deps). See geosaurus/issues/964
    "install_requires": dependencies,
    "setup_requires": dependencies,
    # These classes will execute code after 'pip install' finishes
    # In this case, it will activate the 'arcgis' ipywidget
    # See the top of this setup.py file
    "cmdclass": {
        "develop": develop,
        "install": install,
        "egg_info": egg_info,
    },
    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    "extras_require": {
        "gp": ["dill"],
    },
    "distclass": BinaryDistribution,
    # extras_require={
    #     'dev': ['check-manifest'],
    #     'test': ['coverage'],
    # },
    # If there are data files included in your packages that need to be
    # installed, specify them here.  If using Python 2.6 or less, then these
    # have to be included in MANIFEST.in as well.
    # package_data={
    #     'sample': ['package_data.dat'],
    # },
    "package_data": {
        "arcgis": [
            "gis/_impl/*.pyd",
            "gis/_impl/*.so",
            "graph/_decoder/**/*.pyd",
            "graph/_decoder/**/*.so",
            "learn/*.dll",
            "learn/*.so",
            "learn/_mmdetection_config/*.py",
            "learn/_mmdetection_config/**/*.py",
            "learn/_mmdetection_config/**/**/*.py",
            "learn/_mmseg_config/*.py",
            "learn/_tracking/*.pyd",
            "learn/_tracking/*.dll",
            "raster/*.dll",
            "raster/*.so",
        ],
    },
    # Although 'package_data' is the preferred approach, in some case you may
    # need to place data files outside of your packages. See:
    # http://docs.python.org/3.4/distutils/setupscript.html#installing-additional-files # noqa
    # In this case, 'data_file' will be installed into '<sys.prefix>/my_data'
    # data_files=[('my_data', ['data/data_file'])],
    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    # entry_points={
    #     'console_scripts': [
    #         'sample=sample:main',
    #     ],
    # },
}

try:
    setup(**kwargs)
except Exception:
    log.exception(
        "Exception encountered when attempting to install: Setting "
        "`setup_requires` arg to `[]` and trying again. Exception:\n-----\n"
    )
    kwargs["setup_requires"] = []
    setup(**kwargs)



from ._multi_task_road_extractor import MultiTaskRoadExtractor


class ConnectNet(MultiTaskRoadExtractor):
    """
    Creates a ConnectNet model for binary segmentation of linear features. Supports RGB
    and Multispectral Imagery.
    Implementation based on https://doi.org/10.1109/CVPR.2019.01063 .

    =====================   =====================================================
    **Parameter**            **Description**
    ---------------------   -----------------------------------------------------
    data                    Required fastai Databunch. Returned data object from
                            :meth:`~arcgis.learn.prepare_data`  function.
    ---------------------   -----------------------------------------------------
    backbone                Optional String. Backbone CNN model to be used for
                            creating the base. If hourglass is chosen as
                            the `mtl_model` (Architecture), then this parameter
                            is ignored as hourglass uses a special customised
                            architecture.
                            This parameter is to be used with
                            `linknet` architecture.
                            Default: 'resnet34'

                            Use `supported_backbones` property to get the list
                            of all the supported backbones.
    ---------------------   -----------------------------------------------------
    pretrained_path         Optional String. Path where a compatible pre-trained
                            model is saved. Accepts a Deep Learning Package
                            (DLPK) or Esri Model Definition(EMD) file.
    =====================   =====================================================

    **kwargs**

    =============================   =============================================
    **Parameter**                    **Description**
    -----------------------------   ---------------------------------------------
    mtl_model                       Optional String. It is used to create model
                                    from linknet or
                                    hourglass based neural architectures.
                                    Supported: 'linknet', 'hourglass'.
                                    Default: 'hourglass'
    -----------------------------   ---------------------------------------------
    gaussian_thresh                 Optional float. Sets the gaussian threshold
                                    which allows to set the required width of
                                    the linear feature.
                                    Range: 0.0 to 1.0
                                    Default: 0.76
    -----------------------------   ---------------------------------------------
    orient_bin_size                 Optional Int. Sets the bin size for
                                    orientation angles.
                                    Default: 20
    -----------------------------   ---------------------------------------------
    orient_theta                    Optional Int. Sets the width of orientation
                                    mask.
                                    Default: 8
    =============================   =============================================

    :return: :class:`~arcgis.learn.ConnectNet` Object
    """

    pass
    # This is a dummy class just to hold the doc strings of the model ConnectNet
    # This model is a clone of MultiTaskRoadExtractor and hence not reimplemented.
